
/*****
      eDi_gumbel.c -
      Calculate the expected spacing i of n points drawn from a Gumbel
      distribution with scale s (defaults to 1) and print to stdout.  If
      i is omitted or invalid (2 t/m n) run all over the valid range.
      -trace prints the series by term, -paired runs the v4 sum which
      pairs terms with the same combinatorial factor. -dq uses the
      derivative of the inverse cdf, and -compare prints both the 
      analytic and dQ values and their differences.  -inner evaluates on
      the inner integral (the fDi density) at spacing y.  Uses the mpfr
      package for high precision arithmetic.

      Compile with
        gcc -o eDi_gumbel eDi_gumbel.c -lmpfr

      Call:
        eDi_gumbel [i] n [s] [y] [-trace] [-paired|-dq|-compare|-inner]

      Permission is hereby granted, free of charge, to any person obtaining a
      copy of this software, to deal in the software without restriction,
      including without limitation the rights to use, copy, modify, merge,
      publish, distribute, sublicense, and/or sell copies of the software, and
      to permit persons to whom the software is furnished to do so, subject
      to the following conditions:  The copyright notice and this permission
      notice shall be included in all copies or substantial portions of the
      software;  The software is provided "as is", without warranty of any 
      kind, express or implied, including but no limited to the warranties of
      merchantability, fitness for a particular purpose, and noninfringement;
      In no event shall the authors or copyright holders be liable for any
      claim, damages, or other liability, whether in action of contract, tort
      or otherwise, arising from, out of, or in connection with the software
      or the use or other dealings in the software.

      c 2021 Primordial Machine Vision Systems
*****/

/**
   Notes:
   - By delaying the division for the combinatorial when calculating the pdf,
     ie. keeping the numerator and denominator separate and dividing only at
     the end, to get the term, the pdf matches numeric integration.
   - Hmm, still large differences to the numeric integration for n > 30.  Even
     larger than exp.Dgumbel.  Maybe it's not a convergence problem and the
     series solution for the pdf is bad?
   - If we have a huge precision, why is there a difference between the
     whole sum and the paired split?
     - By delaying the division for the combintaorial in the pdf, handling
       the numerator and denominator separately and dividing at the end,
       we match the numeric integration for n=100, varying either y or i.
     - This does not help the expected spacing.
**/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <mpfr.h>


/**** Constants *****/

/* abbreviation for the rounding direction we will always use. */
#define RND             (MPFR_RNDN)

/* numerical precision, something big */
#define NBIT            4096

/* true to delay divisions as long as possible */
#define SEP_NUMDENOM    0

/* what to do */
#define ACT_EDI         0
#define ACT_PAIRED      1
#define ACT_DQ          2
#define ACT_COMPARE     3
#define ACT_INNER       4


/**** Structures ****/

struct cmdargs {
	int i;                                /* spacing index */
	int n;                                /* number of samples */
	double s;                             /* distribution scale */
	double y;                             /* spacing */
	int trace;                            /* true to print series by term */
	int action;                           /* ACT_* of what to do */
};


/**** Usage Text ****/

char *usage_cmdline[] = {
	"eDi_gumbel [i] n [s] [y] [-trace] [-paired|-dq|-compare|-inner]",
	NULL
};


/**** Prototypes ****/

static void exp_Dgumbel(mpfr_t, int, int, double, int);
static void exp_Dgumbel_paired(mpfr_t, int, int, double, int);
static void dq_gumbel(mpfr_t, int, int, double);
static void pdf_Dgumbel(mpfr_t, int, int, double, double, int);
static void mpout(const char *, mpfr_t, size_t);
static void mplogdiv(mpfr_t, int, int);
static void mpchoose(mpfr_t, int, int);
static void mpchoose_nd(mpfr_t, mpfr_t, int, int);

static void usage(const char *, ...);
static void parse_commandline(int, char **, struct cmdargs *);
static int has_helpopt(int, char **);
static int match_arg(const char *, const char *, unsigned int);


/**** Expected Gumbel spacing ****/

/***
    exp_Dgumbel:  Calculate the expected value of the i'th spacing of n points
                  drawn from a Gumbel distribution.  Base version, series does
                  not sum correctly for large n.
    args:         res - expected spacing
                  i - spacing index (2 t/m n)
                  n - number of points drawn
                  s - distribution scale parameter
                  trace - true to print series sum by term
    modifies:  res
***/
static void exp_Dgumbel(mpfr_t res, int i, int n, double s, int trace) {
	mpfr_t term;                          /* high precision series term */
	mpfr_t tot;                           /* high prec series sum */
	mpfr_t num;                           /* numerator of series term */
	mpfr_t denom;                         /* denominator of series term */
	mpfr_t z1;                            /* intermediate values */
	mpfr_t z2;
	int k;
	int j;

	mpfr_init2(term, NBIT);
	mpfr_init2(tot, NBIT);
	mpfr_init2(num, NBIT);
	mpfr_init2(denom, NBIT);
	mpfr_init2(z1, NBIT);
	mpfr_init2(z2, NBIT);
	
	mpfr_log_ui(z1, i-1, RND);
	mpfr_div_si(z2, z1, -(n-i+1), RND);
	mpfr_set(tot, z2, RND);
	if (trace) {
		printf("%4s %3s %18s   %8s   %3s   %15.8e   %15s\n",
		       "init", "sgn", "comb", "log", "mk1", mpfr_get_d(tot, RND), "tot");
	}

	for (k=0; k<=(n-i); k++) {
		/* Doing this without splitting the numerator/denominator results in
		   the same values as the base or paired version. */
		mpchoose_nd(num, denom, n-i, k);
		mpfr_log_ui(z1, i+k, RND);
		if (trace) {
			printf(" %3d  %c  %18.5f * %8.5f", k, ((0 == ((n-i+k)%2)) ? '+' : '-'),
			       mpfr_get_d(num, RND) / mpfr_get_d(denom, RND),
			       mpfr_get_d(z1, RND));
		}
		mpfr_mul(num, num, z1, RND);
		mpfr_mul_si(denom, denom, k+1, RND);
		mpfr_div(term, num, denom, RND);
		if (0 == (k % 2)) {
			mpfr_add(tot, tot, term, RND);
		} else {
			mpfr_sub(tot, tot, term, RND);
		}
		if (trace) {
			printf(" / %3d = %15.8e   %15.8e\n",
			       k+1, mpfr_get_d(term, RND), mpfr_get_d(tot, RND));
		}
	}

	mpchoose(z2, n, i);
	mpfr_mul_si(z1, z2, i, RND);
	mpfr_mul_d(term, z1, s, RND);
	mpfr_mul(res, tot, term, RND);
	if (trace) {
		printf(" %3s %3s %18ld * %8.5f * %3d   %15.8e   %15.8e\n",
		       " ", " ", mpfr_get_si(z2, RND), ((0 == ((n-i)%2)) ? s : -s), i,
		       mpfr_get_d(term, RND), mpfr_get_d(res, RND));
	}

	mpfr_clear(term);
	mpfr_clear(tot);
	mpfr_clear(num);
	mpfr_clear(denom);
	mpfr_clear(z1);
	mpfr_clear(z2);
}

/* Base version of the series.  This generates bad values for reasonable n
   (> 30) and small i.  Splitting the numerator/denominator of the
   combinatorial helps improve accuracy, but the result still breaks down at
   large n, small i pairs. */
static void exp_Dgumbel_v1(mpfr_t res, int i, int n, double s, int trace) {
	mpfr_t term;                          /* high precision series term */
	mpfr_t tot;                           /* high prec series sum */
	mpfr_t x;                             /* factors in calculations */
	mpfr_t y;
	mpfr_t z1;
	mpfr_t z2;
	mpfr_t z3;
	int k;

	mpfr_init2(term, NBIT);
	mpfr_init2(tot, NBIT);
	mpfr_init2(x, NBIT);
	mpfr_init2(y, NBIT);
	mpfr_init2(z1, NBIT);
	mpfr_init2(z2, NBIT);
	mpfr_init2(z3, NBIT);

	mplogdiv(z1, n, i-1);
	mpfr_set_si(x, n-i+1, RND);
	mpfr_div(tot, z1, x, RND);
	
	if (trace) {
		printf("  %3s %3s %12s   %7s   %19s   %19s\n",
		       "k", "sgn", "comb/k", "log", "term", "cum");
		printf("  %3d  %c  %12s   %7.5f   %19.6f\n",
		       0, '+', " ", mpfr_get_d(z1, RND), mpfr_get_d(tot, RND));
	}
	
	for (k=1; k<=(n-i); k++) {
		mplogdiv(z3, n-k, i-1);
#if SEP_NUMDENOM
		mpchoose_nd(z1, z2, n-i, k-1);
		mpfr_mul(z1, z1, z3, RND);
		mpfr_mul_si(z2, z2, k, RND);
		mpfr_div(term, z1, z2, RND);
		/* This is only for the trace. */
		if (trace) {
			mpchoose(z2, n-i, k-1);
			mpfr_div_si(z2, z2, k, RND);
		}
#else
		mpchoose(z1, n-i, k-1);
		mpfr_set_si(x, k, RND);
		mpfr_div(z2, z1, x, RND);
		mpfr_mul(term, z2, z3, RND);
#endif
		if (0 == (k % 2)) {
  		mpfr_add(tot, tot, term, RND);
		} else {
			mpfr_sub(tot, tot, term, RND);
		}
		if (trace) {
			printf("  %3d  %c  %12ld * %7.5f = %19.6f   %19.6f\n",
			       k, ((0 == (k % 2)) ? '+' : '-'), mpfr_get_si(z2, RND),
			       mpfr_get_d(z3, RND), mpfr_get_d(term, RND), mpfr_get_d(tot, RND));
		}
	}

#if SEP_NUMDENOM
	mpchoose_nd(z1, z2, n, i);
	if (0 == ((n - i) % 2)) {
		mpfr_mul_si(z1, z1, i, RND);
	} else {
		mpfr_mul_si(z1, z1, -i, RND);
	}
	mpfr_mul_d(z2, z2, s, RND);
	mpfr_div(z3, z1, z2, RND);
#else
	mpchoose(z1, n, i);
	if (0 == ((n - i) % 2)) {
		mpfr_set_si(x, i, RND);
	} else {
		mpfr_set_si(x, -i, RND);
	}
	mpfr_mul(z2, z1, x, RND);
	mpfr_div_d(z3, z2, s, RND);
#endif
	if (trace) {
		printf("  sum        %12.5e\n", mpfr_get_d(tot, RND));
		printf("  pre-mult   %12.5e\n", mpfr_get_d(z3, RND));
	}
	
	mpfr_mul(res, z3, tot, RND);

	mpfr_clear(term);
	mpfr_clear(tot);
	mpfr_clear(x);
	mpfr_clear(y);
	mpfr_clear(z1);
	mpfr_clear(z2);
	mpfr_clear(z3);
	mpfr_free_cache();
}

/* This variant moves the (n i) combinatorial inside the series and handles
   its numerator/denominator separately.  These two are very large (vs. the
   terms in the base version, which are similar in size) and yield the same
   result. */
static void exp_Dgumbel_v2(mpfr_t res, int i, int n, double s, int trace) {
	mpfr_t term;                          /* high precision series term */
	mpfr_t tot;                           /* high prec series sum */
	mpfr_t num;                           /* numerator of series term */
	mpfr_t denom;                         /* denominator of series term */
	mpfr_t z1;                            /* intermediate values */
	mpfr_t z2;
	int k;
	int j;

	mpfr_init2(term, NBIT);
	mpfr_init2(tot, NBIT);
	mpfr_init2(num, NBIT);
	mpfr_init2(denom, NBIT);
	mpfr_init2(z1, NBIT);
	mpfr_init2(z2, NBIT);

	if (trace) {
		printf(" %3s %3s %12s   %8s   %11s   %11s   %15s\n",
		       "k", "sgn", "num", "ln", "denom", "term", "cum");
	}

	mpfr_set_si(tot, 0, RND);
	for (k=0; k<=(n-i); k++) {
		mpfr_set_si(num, 1, RND);
		mpfr_set_si(denom, 1, RND);
		for (j=n-i-k+2; j<=n; j++) {
			mpfr_mul_si(num, num, j, RND);
		}
		for (j=2; j<=(i-1); j++) {
			mpfr_mul_si(denom, denom, j, RND);
		}
		for (j=2; j<=k; j++) {
			mpfr_mul_si(denom, denom, j, RND);
		}
		mplogdiv(z1, n-k, i-1);
		mpfr_mul(z2, num, z1, RND);
		mpfr_div(term, z2, denom, RND);
		if (0 == (k % 2)) {
			mpfr_add(tot, tot, term, RND);
		} else {
			mpfr_sub(tot, tot, term, RND);
		}
		if (trace) {
			printf(" %3d  %c  %12.5e * %8.5f / %11.5e = %11.5e   %15.8e\n",
			       k, ((0 == (k % 2)) ? '+' : '-'), mpfr_get_d(num, RND),
			       mpfr_get_d(z1, RND), mpfr_get_d(denom, RND),
			       mpfr_get_d(term, RND), mpfr_get_d(tot, RND));
		}
	}
	
	if (0 == ((n-i) % 2)) {
		mpfr_mul_d(res, tot, s, RND);
	} else {
		mpfr_mul_d(res, tot, -s, RND);
	}
	if (trace) {
		printf(" %3s     %12.5e * %8.5f   %11s = %11s   %11.8f\n",
		       "fin", mpfr_get_d(tot, RND), ((0 == ((n-i) % 2)) ? s : -s),
		       " ", " ", mpfr_get_d(res, RND));
	}

	mpfr_clear(term);
	mpfr_clear(tot);
	mpfr_clear(num);
	mpfr_clear(denom);
	mpfr_clear(z1);
	mpfr_clear(z2);
}

/***
    exp_Dgumbel_paired  Variant 4, dividing the series into paired terms with
                        the sae combinatorial factor (with the pairs different
                        for odd/even n-i).
    args:               res - expected spacing
                        i - spacing index (2 t/m n)
                        n - number of points drawn
                        s - distribution scale parameter
                        trace - true to print series by term
    modifies:  res
***/
static void exp_Dgumbel_paired(mpfr_t res, int i, int n, double s, int trace) {
	mpfr_t term;                          /* high precision series term */
	mpfr_t tot;                           /* high prec series sum */
	mpfr_t x;                             /* factors in calculations */
	mpfr_t y;
	mpfr_t z1;
	mpfr_t z2;
	mpfr_t z3;
	double l1, l2;                        /* intermediate values for tracing */
	int k;

	mpfr_init2(term, NBIT);
	mpfr_init2(tot, NBIT);
	mpfr_init2(x, NBIT);
	mpfr_init2(y, NBIT);
	mpfr_init2(z1, NBIT);
	mpfr_init2(z2, NBIT);
	mpfr_init2(z3, NBIT);

	mpfr_set_d(tot, 0., RND);
	
	if (0 == ((n - i) % 2)) {
		/* The R version tests i < n, but that's because it will generate a
		   sequence in the negative direction, but the for will not iterate. */
		for (k=0; k<=((n-i-2)/2); k++) {
			mplogdiv(z1, n-k, i-1);
			mpfr_set_si(x, n-i+1-k, RND);
			mpfr_div(z2, z1, x, RND);
			
			mplogdiv(z1, i+k, i-1);
			mpfr_set_si(x, k+1, RND);
			mpfr_div(z3, z1, x, RND);

			if (trace) {
				l1 = mpfr_get_d(z2, RND);
				l2 = mpfr_get_d(z3, RND);
			}

			mpfr_add(z1, z2, z3, RND);
				
			mpchoose(z2, n-i, k);
			mpfr_mul(term, z1, z2, RND);
			if (0 == (k % 2)) {
				mpfr_add(tot, tot, term, RND);
			} else {
				mpfr_sub(tot, tot, term, RND);
			}
			if (trace) {
				printf("  %2d e   %c  %20.2f * (%7.5f + %7.5f) = %15.8e\n",
				       k, (0 == (k % 2) ? '+' : '-'),
				       mpfr_get_d(z2, RND), l1, l2, mpfr_get_d(term, RND));
			}
		}
		
		mpchoose(z1, n-i, (n-i)/2);
		if (trace) {
			l1 = mpfr_get_d(z1, RND);
		}
		mpfr_set_si(x, 2, RND);
		mpfr_mul(z2, z1, x, RND);
		mpfr_set_si(x, n-i+2, RND);
		mpfr_div(z3, z2, x, RND);
		mplogdiv(z1, n+i, 2*(i-1));
		mpfr_mul(term, z1, z3, RND);
		if (0 == (((n - i) / 2) % 2)) {
			mpfr_add(tot, tot, term, RND);
		} else {
			mpfr_sub(tot, tot, term, RND);
		}
		if (trace) {
			printf(" mid e   %c  %20.2f *  %7.5f   %7s  = %15.8e\n",
			       ((0 == (((n-i)/2) % 2)) ? '+' : '-'), l1,
			       mpfr_get_d(z1, RND), " ", mpfr_get_d(term, RND));
		}
	} else {
		for (k=0; k<=((n-i-1)/2); k++) {
			mplogdiv(z1, n-k, i-1);
			mpfr_set_si(x, n-i+1-k, RND);
			mpfr_div(z2, z1, x, RND);
			
			mplogdiv(z1, i+k, i-1);
			mpfr_set_si(x, k+1, RND);
			mpfr_div(z3, z1, x, RND);

			if (trace) {
				l1 = mpfr_get_d(z2, RND);
				l2 = mpfr_get_d(z3, RND);
			}

			mpfr_sub(z1, z2, z3, RND);

			mpchoose(z2, n-i, k);
			mpfr_mul(term, z1, z2, RND);
			if (0 == (k % 2)) {
				mpfr_add(tot, tot, term, RND);
			} else {
				mpfr_sub(tot, tot, term, RND);
			}
			if (trace) {
				printf("  %2d o   %c  %20.2f * (%7.5f - %7.5f) = %15.8e\n",
				       k, (0 == (k % 2) ? '+' : '-'),
				       mpfr_get_d(z2, RND), l1, l2, mpfr_get_d(term, RND));
			}
		}
	}

	mpchoose(z1, n, i);
	if (0 == ((n - i) % 2)) {
		mpfr_set_si(x, i, RND);
	} else {
		mpfr_set_si(x, -i, RND);
	}
	mpfr_mul(z2, z1, x, RND);
	mpfr_div_d(z3, z2, s, RND);
	if (trace) {
		printf("  sum        %12.5e\n", mpfr_get_d(tot, RND));
		printf("  pre-mult   %12.5e\n", mpfr_get_d(z3, RND));
	}
	
	mpfr_mul(res, z3, tot, RND);

	mpfr_clear(term);
	mpfr_clear(tot);
	mpfr_clear(x);
	mpfr_clear(y);
	mpfr_clear(z1);
	mpfr_clear(z2);
	mpfr_clear(z3);
	mpfr_free_cache();
}

/***
    dq_gumbel:  Calculate the expected i'th spacing (of n points drawn from a
                Gumbel distribution) using the derivative of the quantile.
    args:       res - expected spacing
                i - spacing index
                n - number of samples drawn
                s - distribution scale parameter
    modifies:  res
***/
static void dq_gumbel(mpfr_t res, int i, int n, double s) {
	mpfr_t p;
	mpfr_t dp;
	mpfr_t u;
	mpfr_t v;

	mpfr_init2(p, NBIT);
	mpfr_init2(dp, NBIT);
	mpfr_init2(u, NBIT);
	mpfr_init2(v, NBIT);

	/* p = (i-1)/n, -dp=-1/n */
	mpfr_set_si(u, i-1, RND);
	mpfr_set_si(v, n, RND);
	mpfr_div(p, u, v, RND);
	mpfr_d_div(dp, -1.0, v, RND);

	/* 1 / (p * log(p)) */
	mpfr_log(u, p, RND);
	mpfr_mul(v, u, p, RND);
	mpfr_d_div(u, 1.0, v, RND);

	/* -dp * s / (p * log(p)) */
	mpfr_mul_d(v, u, s, RND);
	mpfr_mul(res, v, dp, RND);

	mpfr_clear(p);
	mpfr_clear(dp);
	mpfr_clear(u);
	mpfr_clear(v);
}

/***
    pdf_Dgumbel:  Calculate the inner integral for the expected spacing, which
                  is the density of the spacing (fDi).
    args:         res - spacing density
                  i - spacing index (2 t/m n)
                  n - number of points drawn
                  s - distribution scale parameter
                  y - spacing to calculate density at
                  trace - true to print series sum by term
    modifies:  res
***/
static void pdf_Dgumbel(mpfr_t res, int i, int n, double s, double y,
                        int trace) {
	mpfr_t term;                          /* high precision series term */
	mpfr_t tot;                           /* series total */
	mpfr_t u;                             /* intermediate values */
	mpfr_t v;                             /* intermediate values */
	mpfr_t z1;                            /* intermediate values */
	mpfr_t z2;                            /* intermediate values */
	mpfr_t A;                             /* exp(-y/s) */
	mpfr_t denom;                         /* denominator of series term */
	int k;

	mpfr_init2(term, NBIT);
	mpfr_init2(tot, NBIT);
	mpfr_init2(u, NBIT);
	mpfr_init2(v, NBIT);
	mpfr_init2(z1, NBIT);
	mpfr_init2(z2, NBIT);
	mpfr_init2(A, NBIT);
	mpfr_init2(denom, NBIT);

	mpfr_set_d(tot, 0, RND);
	
	/* Denominator independent of k, ie. exp(y/s) * (i-1) + (n - i + 1). */
	mpfr_set_d(u, y, RND);
	mpfr_div_d(v, u, s, RND);
	mpfr_exp(A, v, RND);
	mpfr_mul_si(z1, A, i-1, RND);
	mpfr_add_si(denom, z1, n-i+1, RND);
	for (k=0; k<=(n-i); k++) {
		/* 1 / (denom - k)^2 */
		mpfr_sub_si(z1, denom, k, RND);
		mpfr_sqr(z2, z1, RND);
#if SEP_NUMDENOM
		mpchoose_nd(u, v, n-i, k);
		mpfr_mul(z2, z2, v, RND);
		mpfr_div(term, u, z2, RND);
		if (trace) {
			mpfr_div(z1, u, v, RND);
		}
#else
		/* (n-i i) / (denom - k)^2 */
		mpchoose(z1, n-i, k);
		mpfr_div(term, z1, z2, RND);
#endif
		if (0 == (k % 2)) {
			mpfr_add(tot, tot, term, RND);
		} else {
			mpfr_sub(tot, tot, term, RND);
		}
		if (trace) {
			printf("k %3d  %c  %14ld / %14.6e = %14.6f   %14.6e\n",
			       k, ((0 == (k % 2)) ? '+' : '-'), mpfr_get_si(z1, RND),
			       mpfr_get_d(z2, RND), mpfr_get_d(term, RND), mpfr_get_d(tot, RND));
		}
	}

	/* n!/(i-2)!(n-i)! via x!=loggamma(x+1) and exp. */
	/*
	mpfr_set_si(u, n+1, RND);
	mpfr_lngamma(z1, u, RND);
	mpfr_set_si(u, i-2+1, RND);
	mpfr_lngamma(z2, u, RND);
	mpfr_sub(v, z1, z2, RND);
	mpfr_set_si(u, n-i+1, RND);
	mpfr_lngamma(z1, u, RND);
	mpfr_sub(u, v, z1, RND);
	mpfr_exp(z1, u, RND);
	*/

	/* n! / (i-2)! (n-i)!  by canceling largest denom from num, then dividing
	   by other. */
	mpfr_set_si(u, 1, RND);
	mpfr_set_si(v, 1, RND);
	if ((i - 2) < (n - i)) {
		for (k=n-i+1; k<=n; k++) {
			mpfr_mul_si(u, u, k, RND);
		}
		for (k=2; k<=(i-2); k++) {
			mpfr_mul_si(v, v, k, RND);
		}
	} else {
		for (k=i-1; k<=n; k++) {
			mpfr_mul_si(u, u, k, RND);
		}
		for (k=2; k<=(n-i); k++) {
			mpfr_mul_si(v, v, k, RND);
		}
	}
	mpfr_div(z1, u, v, RND);
	/*
	*/
	
	mpfr_mul(z2, z1, A, RND);
	if (0 != ((n-i) % 2)) {
		mpfr_mul_si(z2, z2, -1, RND);
	}
	if (trace) {
		printf("  sum      %14.5e\n", mpfr_get_d(tot, RND));
		printf("  pre   %c  %14.5e * %14.6f = %14.6e\n",
		       ((0 != ((n-i) % 2)) ? '-' : '+'),
		       mpfr_get_d(z1, RND), mpfr_get_d(A, RND), mpfr_get_d(z2, RND));
	}

	mpfr_mul(res, z2, tot, RND);

	mpfr_clear(term);
	mpfr_clear(tot);
	mpfr_clear(u);
	mpfr_clear(v);
	mpfr_clear(z1);
	mpfr_clear(z2);
	mpfr_clear(A);
	mpfr_clear(denom);
}

/***
    mpout:  Output a high precision variable, possibly preceding with a text.
    args:   pre - text to print before number (skip if NULL or empty)
            v - value to print
            ndec - number of decimal places
***/
static void mpout(const char *pre, mpfr_t v, size_t ndec) {

	if ((NULL != pre) && (1 <= strlen(pre))) {
		printf(pre);
	}
	mpfr_out_str(stdout, 10, ndec, v, RND);
}

/***
    mplogdiv:  Calculate log(num/denom).
    args:      dest - pre-allocated storage for result
               num - numerator
               denom - denominator
    modifies:  dest
***/
static void mplogdiv(mpfr_t dest, int num, int denom) {
	mpfr_t x;                             /* high precision numerator */
	mpfr_t y;                             /* high prec denominator */
	mpfr_t xy;                            /* num/denom */

	mpfr_init(x);
	mpfr_init(y);
	mpfr_init(xy);

	mpfr_set_si(x, num, RND);
	mpfr_set_si(y, denom, RND);
	mpfr_div(xy, x, y, RND);
	mpfr_log(dest, xy, RND);

	mpfr_clear(x);
	mpfr_clear(y);
	mpfr_clear(xy);
}

/* This avoids the division by calculating ln(num) - ln(denom).  It gives
   larger sums. */
static void mplogdiv_v2(mpfr_t dest, int num, int denom) {
	mpfr_t x;                             /* high precision numerator */
	mpfr_t y;                             /* high prec denominator */
	mpfr_t xln;                           /* ln of x */
	mpfr_t yln;                           /* ln of y */

	mpfr_init(x);
	mpfr_init(y);
	mpfr_init(xln);
	mpfr_init(yln);

	/* ln(num/denom) = ln(num) - ln(denom) */
	mpfr_set_si(x, num, RND);
	mpfr_set_si(y, denom, RND);
	mpfr_log(xln, x, RND);
	mpfr_log(yln, y, RND);
	mpfr_sub(dest, xln, yln, RND);

	mpfr_clear(x);
	mpfr_clear(y);
	mpfr_clear(xln);
	mpfr_clear(yln);
}

/***
    mpchoose:  Calculate the combinatorial (a b) = a! / (b! (a-b)!), ie. the
               number of combinations of a elements taken b at a time.
    args:      dest - pre-allocated storage for result
               a, b - sample sizes
    modifies:  dest
***/
static void mpchoose(mpfr_t dest, int a, int b) {
	mpfr_t num;                           /* product in numerator */
	mpfr_t denom;                         /* factorial in denominator */
	int i;

	mpfr_init(num);
	mpfr_init(denom);

	mpchoose_nd(num, denom, a, b);

	mpfr_div(dest, num, denom, RND);

	mpfr_clear(num);
	mpfr_clear(denom);
}

/***
    mpchoose_nd:  Partial evaluation of the combinatorial (a b) = a!/(b! (a-b)!)
                  calculating the numerator and denominator but not doing the
                  division.
    args:         num, denom - pre-allocated combinatorial numerator/denominator
                  a, b - sample sizes
    modifies:  num, denom
***/
static void mpchoose_nd(mpfr_t num, mpfr_t denom, int a, int b) {
	int i;

	mpfr_set_si(num, 1, RND);
	mpfr_set_si(denom, 1, RND);

	/* Make smallest numerator possible. */
	if (b <= (a - b)) {
		for (i=a-b+1; i<=a; i++) {
			mpfr_mul_si(num, num, i, RND);
		}
		for (i=2; i<=b; i++) {
			mpfr_mul_si(denom, denom, i, RND);
		}
	} else {
		for (i=b+1; i<=a; i++) {
			mpfr_mul_si(num, num, i, RND);
		}
		for (i=2; i<=(a-b); i++) {
			mpfr_mul_si(denom, denom, i, RND);
		}
	}
}


/* This uses lgamma to calculate the factorials, adds/subtracts the three,
   and converts back via exp.  Although when printed as an integer the
   two versions match, the sum is smaller with the integer (factorial)
   rather than this. */
static void mpchoose_v2(mpfr_t dest, int a, int b) {
	mpfr_t c;                             /* high precision integer */
	mpfr_t f1;                            /* high prec factorial/intermediary */
	mpfr_t f2;                            /* high prec factorial/intermediary */
	mpfr_t f3;                            /* high prec factorial/intermediary */
	int gsgn;                             /* sign of gamma functions */

	if (a < b) {
		gsgn = a;
		a = b;
		b = gsgn;
	}
	
	if (b < 1) {
		mpfr_set_si(dest, 1, RND);
	} else {
		mpfr_init(c);
		mpfr_init(f1);
		mpfr_init(f2);
		mpfr_init(f3);

		mpfr_set_si(c, b+1, RND);
		mpfr_lgamma(f1, &gsgn, c, RND);
		mpfr_set_si(c, a-b+1, RND);
		mpfr_lgamma(f2, &gsgn, c, RND);
		mpfr_add(f3, f1, f2, RND);

		mpfr_set_si(c, a+1, RND);
		mpfr_lgamma(f1, &gsgn, c, RND);
		mpfr_sub(f2, f1, f3, RND);

		mpfr_exp(dest, f2, RND);
	
		mpfr_clear(c);
		mpfr_clear(f1);
		mpfr_clear(f2);
		mpfr_clear(f3);
	}
}




/**** Top Level ****/

/***
    usage:  Print help message and exit.  The message will be tterminated with
            a newline if one isn't provided.  Requires the command line
            be stored in a global array of strings.
    args:   msg - reason for usage, may be NULL, treated as printf fmt string
            ... - arguments for fmt
***/
static void usage(const char *msg, ...) {
	va_list args;                         /* extra arguments for printf */
	int i;

	if ((NULL != msg) && (0 < strlen(msg))) {
		printf("\nnote: ");
		va_start(args, msg);
		vprintf(msg, args);
		if ('\n' != msg[strlen(msg)-1]) {
			printf("\n");
		}
	}

	printf("\n");
	for (i=0; NULL != usage_cmdline[i]; i++) {
		printf("%6s   ", (0 == i) ? "Usage:" : "");
		printf("%s\n", usage_cmdline[i]);
	}
	printf("\n");		exit(1);
}

/***
    parse_commandline:  Handle the command line.  Exit through usage() if
                        needed.
    args:               argc - number of arguments
                        argv - the arguments
                        args - parsed command line
    modifies:  args
***/
static void parse_commandline(int argc, char **argv, struct cmdargs *args) {
	int nnum;                             /* number of numeric args parsed */
	int i;

	memset(args, 0, sizeof(*args));
	args->action = ACT_EDI;
	args->s = 1.0;
	nnum = 0;

	if (argc < 2) {
		usage(NULL);
	}
	if (has_helpopt(argc, argv)) {
		usage(NULL);
	}
	
	for (i=1; i<argc; i++) {
		if (match_arg(argv[i], "-compare", 2)) {
			args->action = ACT_COMPARE;
		} else if (match_arg(argv[i], "-dq", 2) || match_arg(argv[i], "-dQ", 2)) {
			args->action = ACT_DQ;
		} else if (match_arg(argv[i], "-inner", 2)) {
			args->action = ACT_INNER;
		} else if (match_arg(argv[i], "-paired", 2)) {
			args->action = ACT_PAIRED;
		} else if (match_arg(argv[i], "-trace", 2)) {
			args->trace = 1;
		} else {
			nnum += 1;
			switch (nnum) {
			case 1:
				if (!strcmp(argv[i], "-")) {
					continue;
				}
				args->i = atoi(argv[i]);
				break;
			case 2:
				if (!strcmp(argv[i], "-")) {
					continue;
				}
				args->n = atoi(argv[i]);
				break;
			case 3:
				if (!strcmp(argv[i], "-")) {
					continue;
				}
				args->s = atof(argv[i]);
				break;
			case 4:
				if (!strcmp(argv[i], "-")) {
					continue;
				}
				args->y = atof(argv[i]);
				break;
			default:
				usage("unrecognized argument '%s'", argv[i]);
			}
		}
	}

	if (0 == nnum) {
		usage("must supply n");
	}
	if (1 == nnum) {
		args->n = args->i;
		args->i = 0;
	}
	
	if (args->n < 2) {
		usage("bad n");
	}
	if (args->s <= 0.0) {
		usage("s must be positive");
	}
	if (args->y < 0.0) {
		usage("y must be non-negative");
	}
	if ((args->i < 2) || (args->n < args->i)) {
		args->i = 0;
	}
}

/***
    has_helpopt:  Test if the first command line option is "-h" or "-help"
                  (normally you would call usage(NULL) to print the command
                  line).
    args:         argc - number of command line arguments
                  argv - the arguments
    returns:   true if the first option is "-h" or "-help", 0 if not (no error)
***/
static int has_helpopt(int argc, char **argv) {
	int err;

	if (1 < argc) {
		return !strcmp(argv[1], "-h") || !strcmp(argv[1], "-help");
	}
	return 0;
}

/***
   match_arg:  Test if two strings are equal, matching all of str1 or at
               least nchar characters.
   args:       str1, str2 - strings to compare
               nchar - minimum length of string match
   returns:  true if strings the same, 0 if not
***/
static int match_arg(const char *str1, const char *str2, unsigned int nchar) {
	unsigned int len1;                    /* length of str1 */

	len1 = strlen(str1);
	if (len1 < nchar) {
		return 0;
	}
	if (strlen(str2) < len1) {
		return 0;
	}
	return !strncmp(str1, str2, len1);
}


int main(int argc, char **argv) {
	struct cmdargs args;                  /* command line arguments */
	mpfr_t eDi;                           /* expected spacing */
	mpfr_t eDi2;                          /* expected spacing for compare */
	mpfr_t diff;                          /* eDi - eDi2 */
	int ilo, ihi;                         /* index bounds for spacing */
	int i;
	int err;

	parse_commandline(argc, argv, &args);

	mpfr_init2(eDi, NBIT);
	mpfr_init2(eDi2, NBIT);
	mpfr_init2(diff, NBIT);

	if ((2 <= args.i) && (args.i <= args.n)) {
		ilo = args.i;
		ihi = args.i;
	} else {
		ilo = 2;
		ihi = args.n;
	}

	for (i=ilo; i<=ihi; i++) {
		switch (args.action) {
		case ACT_EDI:
			exp_Dgumbel(eDi, i, args.n, args.s, args.trace);
			break;
		case ACT_PAIRED:
			exp_Dgumbel_paired(eDi, i, args.n, args.s, args.trace);
			break;
		case ACT_DQ:
			dq_gumbel(eDi, i, args.n, args.s);
			break;
		case ACT_INNER:
			pdf_Dgumbel(eDi, i, args.n, args.s, args.y, args.trace);
			break;
		case ACT_COMPARE:
			exp_Dgumbel(eDi, i, args.n, args.s, args.trace);
			dq_gumbel(eDi2, i, args.n, args.s);
			mpfr_sub(diff, eDi, eDi2, RND);
			mpfr_abs(diff, diff, RND);
			printf("  %3d  %18.10g  %18.10g  %18.10g\n",
			       i, mpfr_get_d(eDi, RND), mpfr_get_d(eDi2, RND),
			       mpfr_get_d(diff, RND));
			break;
		default:
			fprintf(stderr, "unknown action");
			err = -1;
			goto cleanup;
		}
		if (ACT_COMPARE != args.action) {
			printf("  %3d  %16.6f\n", i, mpfr_get_d(eDi, RND));
		}
	}

	err = 0;
cleanup:
	mpfr_clear(eDi);
	mpfr_clear(eDi2);
	mpfr_clear(diff);
	return err;
}
