
/*****
      eDi_logistic.c -
      Calculate the expected spacing i of n points drawn from a logistic 
      distribution with scale (stdev) s.  With only one numeric argument
      take n, otherwise fill in left to right.  Values of '-' mean the
      default: i loops for all possible values (2 t/m n), s 1.0, z 0.0.
      -inner evalutes the inner (y) integral at a given z (to verify the
      analytic solution vs. numeric integration), -dq uses the derivative
      of the inverse cdf, and -compare prints both the analytic and
      dQ values and their difference.  -trace prints the series term by
      term.  Uses the mpfr package for high precision arithmetic.

      Compile with
        gcc -o eDi_logistic eDi_logistic.c -lmpfr

      Call:
        eDi_logistic [i|'-'] n [s] [z] [-trace] [-inner|-dq|-compare]

      Permission is hereby granted, free of charge, to any person obtaining a
      copy of this software, to deal in the software without restriction,
      including without limitation the rights to use, copy, modify, merge,
      publish, distribute, sublicense, and/or sell copies of the software, and
      to permit persons to whom the software is furnished to do so, subject
      to the following conditions:  The copyright notice and this permission
      notice shall be included in all copies or substantial portions of the
      software;  The software is provided "as is", without warranty of any 
      kind, express or implied, including but no limited to the warranties of
      merchantability, fitness for a particular purpose, and noninfringement;
      In no event shall the authors or copyright holders be liable for any
      claim, damages, or other liability, whether in action of contract, tort
      or otherwise, arising from, out of, or in connection with the software
      or the use or other dealings in the software.

      c 2021 Primordial Machine Vision Systems
*****/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <mpfr.h>


/**** Constants ****/

/* abbreviation for the rounding direction we will always use. */
#define RND             (MPFR_RNDN)

/* numerical precision, something big */
#define NBIT            4096

/* what to do */
#define ACT_EDI         0
#define ACT_INNER       1
#define ACT_DQ          2
#define ACT_COMPARE     3


/**** Structures ****/

struct cmdargs {
	int i;                                /* spacing index */
	int n;                                /* number of samples */
	double s;                             /* distribution scale */
	double z;                             /* point for inner integral */
	int trace;                            /* true to print series by term */
	int action;                           /* ACT_* of what to do */
};


/**** Usage Text ****/

char *usage_cmdline[] = {
	"eDi_logistic [i] n [s] [z] [-trace] [-inner|-dq|-compare]",
	NULL
};


/**** Prototypes ****/

static void exp_Dlogistic(mpfr_t, int, int, double, int);
static void exp_Dlogistic_inner(mpfr_t, double, int, int, double, int);
static void dq_logistic(mpfr_t, int, int, double);

static void usage(const char *, ...);
static void parse_commandline(int, char **, struct cmdargs *);
static int has_helpopt(int, char **);
static int match_arg(const char *, const char *, unsigned int);


/**** Expected logistic spacing ****/

/***
    exp_Dlogistic:  Expected value of spacing i (= 2 t/m n samples) for a
                    logistic distribution with parameter scale (location
                    ignored).
    args:           res - expected spacing
                    i - spacing index
                    n - number of samples drawn
                    scale - distribution parameter
                    trace - true to print term-by-term values
    modifies:  res
***/
static void exp_Dlogistic(mpfr_t res, int i, int n, double scale, int trace) {
	mpfr_t u;
	mpfr_t v;
	mpfr_t w;
	mpfr_t x;
	mpfr_t term;
	mpfr_t tot;
	int k;

	mpfr_init2(u, NBIT);
	mpfr_init2(v, NBIT);
	mpfr_init2(w, NBIT);
	mpfr_init2(x, NBIT);
	mpfr_init2(term, NBIT);
	mpfr_init2(tot, NBIT);

	mpfr_set_d(tot, 0, RND);

	if (trace && (i < n)) {
		printf("  %3s  %14s / %4s = %14s   %14s\n",
		       "k", "beta", "ni1k", "term", "cum");
	}
	
	/* Note this does nothing if n==i. */
	for (k=1; k<=(n-i); k++) {
		/* B(n-i+2-k, i-1) */
		mpfr_set_si(u, n-i+2-k, RND);
		mpfr_set_si(v, i-1, RND);
		mpfr_beta(w, u, v, RND);
		/* B / (n-i+1-k) */
		mpfr_set_si(u, n-i+1-k, RND);
		mpfr_div(term, w, u, RND);
		mpfr_add(tot, tot, term, RND);
		if (trace) {
			printf("  %3d  %14.7e / %4d = %14.7e   %14.7e\n",
			       k, mpfr_get_d(w, RND), n-i+1-k, mpfr_get_d(term, RND),
			       mpfr_get_d(tot, RND));
		}
	}
	/* 1/(1-i)^2 */
	mpfr_set_si(u, 1-i, RND);
	mpfr_d_div(v, 1.0, u, RND);
	mpfr_mul(term, v, v, RND);
	/* 1/(1-i)^2 - B(n-i+1-k, i-1)/(n-i+1-k) */
	mpfr_sub(tot, term, tot, RND);
	if (trace) {
		printf("  %3d  %14s   %4s   %14.7e   %14.7e\n",
		       0, " ", " ", mpfr_get_d(term, RND), mpfr_get_d(tot, RND));
	}
	
	/* n! / ((i-2)! * (n-i+1)!) via exp, lgamma (n! = gamma(n+1)) */
	mpfr_set_si(u, n+1, RND);
	mpfr_lngamma(v, u, RND);
	mpfr_set_si(u, i-1, RND);
	mpfr_lngamma(w, u, RND);
	mpfr_set_si(u, n-i+2, RND);
	mpfr_lngamma(x, u, RND);
	mpfr_add(u, w, x, RND);
	mpfr_sub(w, v, u, RND);
	mpfr_exp(x, w, RND);

	mpfr_mul_d(v, x, scale, RND);
	mpfr_mul(res, v, tot, RND);
	if (trace) {
		printf("  pre  %14.7e * %4s   %14s   %14.7e\n",
		       mpfr_get_d(v, RND), " ", " ", mpfr_get_d(res, RND));
	}

	mpfr_clear(u);
	mpfr_clear(v);
	mpfr_clear(w);
	mpfr_clear(x);
	mpfr_clear(term);
	mpfr_clear(tot);
}

/***
    exp_Dlogistic_inner:  Calculate the inner integral (over y) of the expected
                          spacing of the i'th point of n drawn from a logistic
                          distribution with scale/stdev s.  Evaluate at the
                          point z.
    args:                 res - inner integral
                          z - point/position
                          i - index of drawn sample
                          n - number of drawn samples
                          s - scale/stdev of distribution
                          trace - true to print series sum by term
    modifies:  res
***/
static void exp_Dlogistic_inner(mpfr_t res, double z, int i, int n, double s,
                                int trace) {
	mpfr_t u;
	mpfr_t v;
	mpfr_t w;
	mpfr_t x;
	mpfr_t y;
	mpfr_t term;
	mpfr_t tot;
	int k;

	mpfr_init2(u, NBIT);
	mpfr_init2(v, NBIT);
	mpfr_init2(w, NBIT);
	mpfr_init2(x, NBIT);
	mpfr_init2(y, NBIT);
	mpfr_init2(term, NBIT);
	mpfr_init2(tot, NBIT);

	mpfr_set_d(tot, 0.0, RND);

	if (trace) {
		printf("  %3s  %12s   %12s / %3s = %12s   %12s\n",
		       "k", "exp", "pwr", "ni1", "term", "cum");
	}
	for (k=1; k<=(n-i); k++) {
		/* y = exp(k*z) */
		mpfr_set_d(u, z, RND);
		mpfr_mul_si(v, u, k, RND);
		mpfr_exp(y, v, RND);
		/* x = (1 / (1 + exp(-z)))^(n-i+1-k) */
		mpfr_set_d(u, -z, RND);
		mpfr_exp(v, u, RND);
		mpfr_add_si(w, v, 1, RND);
		mpfr_si_div(u, 1, w, RND);
		mpfr_set_si(v, n-i+1-k, RND);
		mpfr_pow(x, u, v, RND);
		/* term = x * y / (n - i + 1 - k) */
		mpfr_mul(w, x, y, RND);
		mpfr_div(term, w, v, RND);
		/* cum */
		mpfr_add(tot, tot, term, RND);
		if (trace) {
			printf("  %3d  %12.5e * %12.5e / %3d = %12.5e   %12.5e\n",
			       k, mpfr_get_d(x, RND), mpfr_get_d(y, RND), n-i+1-k,
			       mpfr_get_d(term, RND), mpfr_get_d(tot, RND));
		}
	}

	/* x = exp(z * (n - i + 1)) */
	mpfr_set_si(u, n-i+1, RND);
	mpfr_set_d(v, z, RND);
	mpfr_mul(w, u, v, RND);
	mpfr_exp(x, w, RND);
	/* y = log(1 + exp(-z)) */
	mpfr_set_d(v, -z, RND);
	mpfr_exp(w, v, RND);
	mpfr_add_si(v, w, 1, RND);
	mpfr_log(y, v, RND);
	/* term = x * y, add to total */
	mpfr_mul(term, x, y, RND);
	mpfr_sub(tot, tot, term, RND);
	if (trace) {
		printf("\n");
		printf("  %3s  %12s * %12s   %3s = %12s   %12s\n",
		       " ", "exp", "ln", " ", "term", "pwr-ln");
		printf("  %3s  %12.5e   %12.5e   %3s = %12.5e   %12.5e\n",
		       " ", mpfr_get_d(x, RND), mpfr_get_d(y, RND), " ",
		       mpfr_get_d(term, RND), mpfr_get_d(tot, RND));
	}

	/* res = -s * cum / (n - i + 1) */
	mpfr_d_div(w, -s, u, RND);
	mpfr_mul(res, tot, w, RND);

	mpfr_clear(u);
	mpfr_clear(v);
	mpfr_clear(w);
	mpfr_clear(x);
	mpfr_clear(y);
	mpfr_clear(term);
	mpfr_clear(tot);
}

/***
    dq_logistic:  Calculate the expected i'th spacing (of n points drawn from
                  a logistic distribution with scale) using the derivative
                  of the quantile.
    args:         res - expected spacing
                  i - spacing index
                  n - number of samples drawn
                  scale - distribution parameter
    modifies:  res
***/
static void dq_logistic(mpfr_t res, int i, int n, double scale) {
	mpfr_t p;
	mpfr_t dp;
	mpfr_t u;
	mpfr_t v;

	mpfr_init2(p, NBIT);
	mpfr_init2(dp, NBIT);
	mpfr_init2(u, NBIT);
	mpfr_init2(v, NBIT);

	/* p = (i-1)/n, dp = 1/n */
	mpfr_set_si(u, i-1, RND);
	mpfr_set_si(v, n, RND);
	mpfr_div(p, u, v, RND);
	mpfr_d_div(dp, 1.0, v, RND);
	
	/* 1 / (p * (1 - p)) */
	mpfr_set_si(v, 1, RND);
	mpfr_sub(u, v, p, RND);
	mpfr_mul(v, p, u, RND);
	mpfr_d_div(u, 1.0, v, RND);

	/* scale * dp / (p * (1 - p)) */
	mpfr_mul(v, u, dp, RND);
	mpfr_mul_d(res, v, scale, RND);

	mpfr_clear(p);
	mpfr_clear(dp);
	mpfr_clear(u);
	mpfr_clear(v);
}



/**** Top Level ****/

/***
    usage:  Print help message and exit.  The message will be tterminated with
            a newline if one isn't provided.  Requires the command line
            be stored in a global array of strings.
    args:   msg - reason for usage, may be NULL, treated as printf fmt string
            ... - arguments for fmt
***/
static void usage(const char *msg, ...) {
	va_list args;                         /* extra arguments for printf */
	int i;

	if ((NULL != msg) && (0 < strlen(msg))) {
		printf("\nnote: ");
		va_start(args, msg);
		vprintf(msg, args);
		if ('\n' != msg[strlen(msg)-1]) {
			printf("\n");
		}
	}

	printf("\n");
	for (i=0; NULL != usage_cmdline[i]; i++) {
		printf("%6s   ", (0 == i) ? "Usage:" : "");
		printf("%s\n", usage_cmdline[i]);
	}
	printf("\n");	exit(1);
}

/***
    parse_commandline:  Handle the command line.  Exit through usage() if
                        needed.
    args:               argc - number of arguments
                        argv - the arguments
                        args - parsed command line
    modifies:  args
***/
static void parse_commandline(int argc, char **argv, struct cmdargs *args) {
	int nnum;                             /* number of numeric args parsed */
	int i;

	memset(args, 0, sizeof(*args));
	args->action = ACT_EDI;
	args->s = 1.0;
	nnum = 0;

	if (argc < 2) {
		usage(NULL);
	}
	if (has_helpopt(argc, argv)) {
		usage(NULL);
	}

	for (i=1; i<argc; i++) {
		if (match_arg(argv[i], "-compare", 2)) {
			args->action = ACT_COMPARE;
		} else if (match_arg(argv[i], "-dQ", 2) || match_arg(argv[i], "-dq", 2)) {
			args->action = ACT_DQ;
		} else if (match_arg(argv[i], "-inner", 2)) {
			args->action = ACT_INNER;
		} else if (match_arg(argv[i], "-trace", 2)) {
			args->trace = 1;
		} else {
			nnum += 1;
			switch (nnum) {
			case 1:
				if (!strcmp(argv[i], "-")) {
					continue;
				}
				args->i = atoi(argv[i]);
				break;
			case 2:
				if (!strcmp(argv[i], "-")) {
					continue;
				}
				args->n = atoi(argv[i]);
				break;
			case 3:
				if (!strcmp(argv[i], "-")) {
					continue;
				}
				args->s = atof(argv[i]);
				break;
			case 4:
				if (!strcmp(argv[i], "-")) {
					continue;
				}
				args->z = atof(argv[i]);
				break;
			default:
				usage("unrecognized argument '%s'", argv[i]);
			}
		}
	}

	if (0 == nnum) {
		usage("must supply n");
	}
	if (1 == nnum) {
		args->n = args->i;
		args->i = 0;
	}
	
	if (args->n < 2) {
		usage("n must be >= 2");
	}
	if (args->s <= 0.0) {
		usage("s must be positive");
	}
	if ((args->i < 2) || (args->n < args->i)) {
		args->i = 0;
	}
}

/***
    has_helpopt:  Test if the first command line option is "-h" or "-help"
                  (normally you would call usage(NULL) to print the command
                  line).
    args:         argc - number of command line arguments
                  argv - the arguments
    returns:   true if the first option is "-h" or "-help", 0 if not (no error)
***/
static int has_helpopt(int argc, char **argv) {
	int err;
	if (1 < argc) {
		return !strcmp(argv[1], "-h") || !strcmp(argv[1], "-help");
	}
	return 0;
}

/***
   match_arg:  Test if two strings are equal, matching all of str1 or at
               least nchar characters.
   args:       str1, str2 - strings to compare
               nchar - minimum length of string match
   returns:  true if strings the same, 0 if not
***/
static int match_arg(const char *str1, const char *str2, unsigned int nchar) {
	unsigned int len1;                    /* length of str1 */

	len1 = strlen(str1);
	if (len1 < nchar) {
		return 0;
	}
	if (strlen(str2) < len1) {
		return 0;
	}
	return !strncmp(str1, str2, len1);
}


int main(int argc, char **argv) {
	struct cmdargs args;                  /* command line arguments */
	mpfr_t eDi;                           /* expected spacing */
	mpfr_t eDi2;                          /* expected spacing for compare */
	mpfr_t diff;                          /* eDi - eDi2 */
	int ilo, ihi;                         /* index bounds for spacing */
	int i;
	int err;

	parse_commandline(argc, argv, &args);

	mpfr_init2(eDi, NBIT);
	mpfr_init2(eDi2, NBIT);
	mpfr_init2(diff, NBIT);

	if ((2 <= args.i) && (args.i <= args.n)) {
		ilo = args.i;
		ihi = args.i;
	} else {
		ilo = 2;
		ihi = args.n;
	}

	for (i=ilo; i<=ihi; i++) {
		switch (args.action) {
		case ACT_EDI:
			exp_Dlogistic(eDi, i, args.n, args.s, args.trace);
			break;
		case ACT_INNER:
			exp_Dlogistic_inner(eDi, args.z, i, args.n, args.s, args.trace);
			break;
		case ACT_DQ:
			dq_logistic(eDi, i, args.n, args.s);
			break;
		case ACT_COMPARE:
			exp_Dlogistic(eDi, i, args.n, args.s, args.trace);
			dq_logistic(eDi2, i, args.n, args.s);
			mpfr_sub(diff, eDi, eDi2, RND);
			mpfr_abs(diff, diff, RND);
			printf("  %3d  %18.10g  %18.10g  %18.10g\n",
			       i, mpfr_get_d(eDi, RND), mpfr_get_d(eDi2, RND),
			       mpfr_get_d(diff, RND));
			break;
		default:
			fprintf(stderr, "unknown action");
			err = -1;
			goto cleanup;
		}
		if (ACT_COMPARE != args.action) {
			printf("  %3d  %18.10g\n", i, mpfr_get_d(eDi, RND));
		}
	}

	err = 0;
cleanup:
	mpfr_clear(eDi);
	mpfr_clear(eDi2);
	mpfr_clear(diff);
	return err;
}
