
/*****
      vDi_logistic.c -
      Calculate the variance of the expected spacing i of n points drawn
      from a logistic distribution with scale (stdev) s.  With only one 
      numeric argument take n, otherwise fill in left to right.  Values
      of '-' mean the default: i loops for all possible values (2 t/m n),
      s 1.0, z 0.0.  -inner evalutes the inner (second moment) integral at
      a given z (to verify the analytic solution vs. numeric integration).
      -trace prints information about the calculation, -detail more
      information, possibly term by term.  -bist runs internal tests.

      Uses the mpfr package for high precision arithmetic.

      Compile with
        gcc -o vDi_logistic vDi_logistic.c -lmpfr -lm

      Call:
        vDi_logistic [i|'-'] n [s] [z] [-trace] [-detail] [-inner|-bist]

      Permission is hereby granted, free of charge, to any person obtaining a
      copy of this software, to deal in the software without restriction,
      including without limitation the rights to use, copy, modify, merge,
      publish, distribute, sublicense, and/or sell copies of the software, and
      to permit persons to whom the software is furnished to do so, subject
      to the following conditions:  The copyright notice and this permission
      notice shall be included in all copies or substantial portions of the
      software;  The software is provided "as is", without warranty of any 
      kind, express or implied, including but no limited to the warranties of
      merchantability, fitness for a particular purpose, and noninfringement;
      In no event shall the authors or copyright holders be liable for any
      claim, damages, or other liability, whether in action of contract, tort
      or otherwise, arising from, out of, or in connection with the software
      or the use or other dealings in the software.

      c 2021 Primordial Machine Vision Systems
*****/

/**
   ToDo:
   - Add convergence tolerance (typ. 5e-5) to cut-off dilog sum.  Warn if
     convergence not reached.
     - Convergence should be an order of magnitude or more below the
       pre-scaling factor 2 * scale^2 * n! / ((i-2)! (n-i+1)!).  This is
       an absolute tolerance.  For example, with n=75 the pre-factor is
       1.009e+18.  The increment in out3a (out3a/k^2) drops below at
       k=15700, which is marginally stable.  It drops below 1e-19 at
       k=33700, a little higher than the stable line in the table below.
     - Heuristic: stop when increment is below 10^-ceil(100/n) (0.0001
       for n=25, 0.001 for n=34, 0.01 for n=50, 0.1 for n=75).  The
       smaller sample sizes need higher precision.  For the table below
       we need:
       i = 18  n=25   n=30   n=40   n=50    n=60    n=65    n=70     n=75
       kstable  116    542   2262   5040   17129   28940   75194   > 100k
       but here for n=70 the internal values have overflowed.  Based on
       the table below setting KMAX to 50000.
     - n=30 ; for ((i=2; i<=$n; i++)) ; do echo -n "  i $i:  " ; 
                ./vDi_logistic $i $n -trace | grep converg ; done
       to print how many steps it takes for all possible spacing indices.
   - The Cohen/pracma acceleration algorithm works well.  Max # convergence
     steps over all spacing indices, to 1e-6 of the inverse pre-factor scaling
        n  25    50    75   100  150  200  225  250  300  400  500
           20    30    40    49   69   89   99  108  128  167  207
   - For reasonable n the accuracy depends on KMAX, which needs to be
     large (100 is too small), which costs runtime.  Either figure out
     faster convergence of the out3a, out3b series, or cache the dilog
     out3 integral after calculating it with a huge KMAX (since it
     depends only on i), or get an incremental series as k increases
     for out3[ab].
     - i = 18  n=25     n=30     n=40     n=50     n=75      n=100     n=150
       KMAX    var      var      var      var      var       var
       qev     0.0328   0.0180   0.0102   0.0078   0.0057    0.0050
       50      0.0313  -0.0021  -14      -1601    -4824105  -1081124249
       100     0.0315   0.0161  -0.8968  -101     -303250   -67960979
       250     0.0315   0.0173  -0.0134  -2.5892  -7825     -1753721
       500     0.0315   0.0173   0.0085  -0.1552  -491      -109983
       750     0.0315   0.0173   0.0097  -0.0246  -97       -21752
       1000    0.0315   0.0173   0.0099  -0.0025  -31       -6887
       1250                      0.0100   0.0035  -13       -2822
       1500                      0.0100   0.0056  -6        -1361
       2000                               0.0070  -2        -430
       2500                               0.0074  -0.7894   -176
       3000                               0.0075  -0.3742   -85
       4000                               0.0076  -0.1146   -27
     - incremental version
       5000    0.0315   0.0173   0.0099   0.0076  -0.0436   -11
       7500    0.0315   0.0173   0.0100   0.0076  -0.0041   -2.1753
      10000    0.0315   0.0173   0.0100   0.0077   0.0026   -0.6849
      12500                               0.0077   0.0044   -0.2777
      15000                               0.0077   0.0050   -0.1314
      17500                               0.0077   0.0053   -0.0069
      20000                                        0.0054   -0.0382
      25000                                        0.0056   -0.0127
      30000                                        0.0056   -0.0036
      40000                                        0.0056    0.0022   -4.3729
      50000                                        0.0056    0.0038   -1.7886
      60000                                                  0.0044   -0.8603
      70000                                                  0.0046   -0.4624
      80000                                                  0.0048   -0.2692
      90000                                                  0.0048   -0.1665
     100000                                                  0.0049   -0.1077
     stable       100      250     1250    10000    25000    100000   not
**/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>
#include <mpfr.h>


/**** Constants ****/

/* abbreviation for the rounding direction we will always use. */
#define RND             (MPFR_RNDN)

/* numerical precision, something big */
#define NBIT            4096

/* default maximum dilog series  index if k varying, for variance sum */
#define KMAX            50000

/* convergence terms */
#define NCONV           250

/* fraction of pre-scaling factor to which stop dilog integrals */
#define FSTABLE         1e-6

/* what to do */
#define ACT_VDI         0
#define ACT_INNER       1
#define ACT_BIST        4

/* level of tracing */
#define TRACE_GEN    0x01
#define TRACE_DTL    0x02


/**** Macros ****/

/* return number of elements created with array
     arr - array symbol  (ie. declared arr[] = { .. };
*/
#define LEN(arr)        (sizeof(arr) / sizeof(*(arr)))

/* sign of a term, -1^pwr */
#define SETSIGN(pwr)    ((0 == ((abs(pwr)) % 2)) ? 1 : -1)

/* convert the sign s to a character, '-' or '+' or ' ' for 0 */
#define CSIGN(s)        (((s) < 0) ? '-' : (((s) > 0) ? '+' : ' '))



/**** Structures ****/

struct cmdargs {
	int i;                                /* spacing index */
	int n;                                /* number of samples */
	double s;                             /* distribution scale */
	double z;                             /* point for inner integral */
	int trace;                            /* true to print series by term */
	int cum;                              /* true to sum out3a series terms */
	int action;                           /* ACT_* of what to do */
};

struct increq {
	mpfr_t scl;                           /* scaling of iout equation */
	mpfr_t sum;                           /* sum in iout equation */
	mpfr_t c;                             /* constant in iout equation */
};

/* Allocate temporary variables, and some constants, once.  Is this the
   best way to do it?  (vs. b.v. static vars within functions)  Should
   these just be declared globals rather than passing the structure around?
*/
struct scratch {
	mpfr_t u;                             /* intermediate values */
	mpfr_t v;
	mpfr_t w;
	mpfr_t num;                           /* numerator of a fraction */
	mpfr_t denom;                         /* denominator of a fraction */
	mpfr_t one;                           /* 1 in high precision */
	mpfr_t i1fact;                        /* (i-1)! */
	mpfr_t ifact;                         /* i! */
	int fnum[5];                          /* numerator list for factfrac */
	int fdenom[5];                        /* denominator list for factfrac */
};

struct converge {
	mpfr_t val[NCONV];                    /* cache of values */
	mpfr_t pred[NCONV];                   /* alternating sum */
	mpfr_t b;                             /* acceleration values */
	mpfr_t c;
	mpfr_t d;
	mpfr_t s;
	mpfr_t u;                             /* intermediate values */
	int nconv;                            /* number of values stored */
};
	


/**** Usage Text ****/

char *usage_cmdline[] = {
	"vDi_logistic [i|-] n [s] [z] [-trace] [-detail] [-inner|-bist]",
	NULL
};


/**** Prototypes ****/

static void var_Dlogistic(mpfr_t, int, int, double, int);
static void var_Dlogistic_inner(mpfr_t, double, int, int, double, int);
static void var_Dlogistic_out3a(mpfr_t, struct increq *, struct scratch *,
                                int, int, int);
static void var_Dlogistic_out3b(mpfr_t, struct increq *, struct scratch *,
                                int, int, int);
static void sum_lnconst(mpfr_t, int, struct scratch *);
static void accelerate(struct converge *, mpfr_t, int);
static void simplify_frac(int *, mpfr_t, mpfr_t);
static void simplify_factfrac(mpfr_t, mpfr_t, int *, int, int *, int);
static int sortfn_int(const void *, const void *);
static void remove_nonpos(int *, int *);
static int gcd(int, int);
static void mpfr_gcd(mpfr_t, mpfr_t, mpfr_t);
static void init_scratch(struct scratch *, int);
static void init_incr3(struct increq *);
static void init_converge(struct converge *);
static void clear_scratch(struct scratch *);
static void clear_incr3(struct increq *);
static void clear_converge(struct converge *);

static int BIST_gcd(void);
static int BIST_simplify_factfrac(void);
static int BIST_match_nonpos(int, int *, int, int *, int);
static int BIST_remove_nonpos(void);
static int BIST_init_scratch(void);

static void usage(const char *, ...);
static void parse_commandline(int, char **, struct cmdargs *);
static int has_helpopt(int, char **);
static int match_arg(const char *, const char *, unsigned int);


/**** Expected logistic spacing ****/

/***
    var_Dlogistic:  Variance of the expected value of spacing i (= 2 t/m n
                    samples) for a logistic distribution with parameter scale
                    (location ignored).
    args:           res - spacing variance
                    i - spacing index
                    n - number of samples drawn
                    scale - distribution parameter
                    trace - true to print information about calculation
    modifies:  res
***/
static void var_Dlogistic(mpfr_t res, int i, int n, double scale, int trace) {
	struct increq incr3a;                 /* incremental out3a values */
	struct increq incr3b;                 /* incremental out3b values */
	struct scratch tmp;                   /* intermediate/constant values */
	struct converge sum3a;                /* accelerated out3a sum */
	struct converge sum3b;                /* accelerated out3b sum */
	mpfr_t eDi;                           /* expected spacing */
	mpfr_t eDi2;                          /* second moment */
	mpfr_t out1;                          /* first outer integral */
	mpfr_t out2;                          /* second outer integral */
	mpfr_t out3;                          /* third (dilog) outer integral */
	mpfr_t out3a;                         /* dilog |z| < 1 integral */
	mpfr_t out3b;                         /* dilog |z| > 1 integral */
	mpfr_t out3c;                         /* constant part of dilog inversion */
	mpfr_t out3d;                         /* ln part of dilog inversion */
	mpfr_t pre;                           /* scaling factor outside sum */
	mpfr_t tot;                           /* sum of inner products */
	mpfr_t ksq;                           /* k*k scaling of dilog out3[ab] */
	mpfr_t term;                          /* series term */
	mpfr_t stblim;                        /* stability criterion for out3[ab] */
	int nu;                               /* n - i + 2 abbreviation */
	int sgn;                              /* +/- indicator */
	int stable;                           /* out3[ab] stability flags */
	int k, l, m;

	mpfr_init2(eDi, NBIT);
	mpfr_init2(eDi2, NBIT);
	mpfr_init2(out1, NBIT);
	mpfr_init2(out2, NBIT);
	mpfr_init2(out3, NBIT);
	mpfr_init2(out3a, NBIT);
	mpfr_init2(out3b, NBIT);
	mpfr_init2(out3c, NBIT);
	mpfr_init2(out3d, NBIT);
	mpfr_init2(pre, NBIT);
	mpfr_init2(tot, NBIT);
	mpfr_init2(ksq, NBIT);
	mpfr_init2(term, NBIT);
	mpfr_init2(stblim, NBIT);

	init_scratch(&tmp, i);
	init_incr3(&incr3a);
	init_incr3(&incr3b);
	init_converge(&sum3a);
	init_converge(&sum3b);

	nu = n - i + 2;

	/* pre-scaling = 2 * scale^2 * n! / ((i-2)! (n-i+1)!) */
	tmp.fnum[0] = n;
	tmp.fdenom[0] = i - 2;
	tmp.fdenom[1] = n - i + 1;
	simplify_factfrac(tmp.num, tmp.denom, tmp.fnum, 1, tmp.fdenom, 2);
	mpfr_mul_d(tmp.u, tmp.num, 2*scale, RND);
	mpfr_div(pre, tmp.u, tmp.denom, RND);

	/* out1 = ( \sum 1/(nu-1-k) ) / (-i+1)^2 */
	mpfr_set_zero(out1, 0);
	for (k=1; k<=(nu-2); k++) {
		mpfr_div_si(tmp.u, tmp.one, nu-1-k, RND);
		mpfr_add(out1, out1, tmp.u, RND);
	}
	mpfr_div_si(out1, out1, (-i+1)*(-i+1), RND);

	/* out2 = \sum_k \sum_l (nu-2-k-l)! (i-2)! / ((n-k-l)! (nu-1-k)) */
	mpfr_set_zero(out2, 0);
	for (k=1; k<=(nu-3); k++) {
		for (l=1; l<=(nu-2-k); l++) {
			tmp.fnum[0] = nu - 2 - k - l;
			tmp.fnum[1] = i - 2;
			tmp.fdenom[0] = n - k - l;
			simplify_factfrac(tmp.num, tmp.denom, tmp.fnum, 2, tmp.fdenom, 1);
			mpfr_mul_si(tmp.u, tmp.denom, nu-1-k, RND);
			mpfr_div(tmp.v, tmp.num, tmp.u, RND);
			mpfr_add(out2, out2, tmp.v, RND);
		}
	}

	/* out3a = (-1)^k * ( sum_k out3a ) / k^2
	   out3b = (-1)^k * ( sum_k out3b ) / k^2 */
	sgn = -1;
	stable = 0;
	/* Heuristic that sets tighter convergence for small n. */
	/*
	mpfr_d_div(stblim, pow(10,-ceil(100.0/n)), pre, RND);
	*/
	/* Pick stable limit well below the pre-scaling factor.  We will stop the
	   series acceleration when the delta is less than this, and since that
	   happens quickly we can pick an aggressive limit. */
	mpfr_d_div(stblim, FSTABLE, pre, RND);
	mpfr_set_zero(out3a, 0);
	mpfr_set_zero(out3b, 0);
	for (k=1; (0x03 != stable) && (k<=KMAX); k++, sgn*=-1) {
		mpfr_set_si(ksq, sgn*(k*k), RND);
		
		var_Dlogistic_out3a(term, &incr3a, &tmp, k, i, trace);
		mpfr_div(tmp.u, term, ksq, RND);
		accelerate(&sum3a, tmp.u, 0);
		if (!(stable & 0x01)) {
			if ((1 < k) && (k <= NCONV)) {
				mpfr_sub(tmp.v, sum3a.pred[k-1], sum3a.pred[k-2], RND);
				if (mpfr_cmpabs(tmp.v, stblim) <= 0) {
					stable |= 0x01;
				}
				mpfr_set(out3a, sum3a.pred[k-1], RND);
			} else  {
				mpfr_add(out3a, out3a, tmp.u, RND);
			}
		}
		
		var_Dlogistic_out3b(term, &incr3b, &tmp, k, i, trace);
		mpfr_div(tmp.u, term, ksq, RND);
		accelerate(&sum3b, tmp.u, 0);
		if (!(stable & 0x02)) {
			if ((1 < k) && (k <= NCONV)) {
				mpfr_sub(tmp.v, sum3b.pred[k-1], sum3b.pred[k-2], RND);
				if (mpfr_cmpabs(tmp.v, stblim) <= 0) {
					stable |= 0x02;
				}
				mpfr_set(out3b, sum3b.pred[k-1], RND);
			} else {
				mpfr_add(out3b, out3b, tmp.u, RND);
			}
		}
	}

	if ((TRACE_DTL & trace) && (0x03 == stable)) {
		printf("  dilog out3[ab] series converged at index k = %-3d to limit %13.6e\n", k-1, mpfr_get_d(stblim,RND));
	}
	if (0x03 != stable) {
		printf("  warning: dilog out3[ab] series did not converge\n");
	}

	/* out3c = - pi^2 * 2^{-i+1} / (6 * (-i + 1)) */
	mpfr_const_pi(tmp.u, RND);
	mpfr_sqr(tmp.v, tmp.u, RND);
	mpfr_div_si(tmp.u, tmp.v, 6, RND);
	mpfr_div_si(tmp.v, tmp.u, -i+1, RND);
	mpfr_neg(tmp.w, tmp.v, RND);
	mpfr_mul_2si(out3c, tmp.w, -i+1, RND);

	/* out3d = ( (sum_k (ln 2 - sum_l 1/((i-1-k-l) 2^(i-1-k-l))) / (i-1-k))
	             - (pi^2/12) ) / (-i + 1) */
	mpfr_set_zero(out3d, 0);
	for (k=1; k<=(i-2); k++) {
		mpfr_const_log2(tot, RND);
		for (l=1; l<=(i-2-k); l++) {
			mpfr_div_si(tmp.u, tmp.one, i-1-k-l, RND);
			mpfr_div_2si(tmp.v, tmp.u, i-1-k-l, RND);
			mpfr_sub(tot, tot, tmp.v, RND);
		}
		mpfr_div_si(tmp.u, tot, i-1-k, RND);
		mpfr_add(out3d, out3d, tmp.u, RND);
	}
	mpfr_const_pi(tmp.u, RND);
	mpfr_sqr(tmp.v, tmp.u, RND);
	mpfr_div_si(tmp.u, tmp.v, -12, RND);
	mpfr_add(out3d, out3d, tmp.u, RND);
	mpfr_div_si(out3d, out3d, -i+1, RND);

	/* out3 = out3a - out3b - out3c - out3d */
	mpfr_sub(tmp.u, out3a, out3b, RND);
	mpfr_sub(tmp.v, tmp.u, out3c, RND);
	mpfr_sub(out3, tmp.v, out3d, RND);

	/* second moment = pre * (-out1 + out2 - out3) */
	mpfr_sub(tmp.u, out2, out1, RND);
	mpfr_sub(tmp.v, tmp.u, out3, RND);
	mpfr_mul(eDi2, pre, tmp.v, RND);

	/* expected spacing via quantile deriv
	   eDi = scale * n / ((i - 1) * (n - i + 1)) */
	mpfr_set_si(tmp.u, n, RND);
	mpfr_div_si(tmp.v, tmp.u, i-1, RND);
	mpfr_div_si(tmp.w, tmp.v, n-i+1, RND);
	mpfr_mul_d(eDi, tmp.w, scale, RND);

	/* variance = second moment - expected^2 */
	mpfr_sqr(tmp.u, eDi, RND);
	mpfr_sub(res, eDi2, tmp.u, RND);
	
	if (TRACE_GEN & trace) {
		printf("  dilog    %14.7e - %14.7e - %14.7e - %14.7e\n",
		       mpfr_get_d(out3a, RND), mpfr_get_d(out3b, RND),
		       mpfr_get_d(out3c, RND), mpfr_get_d(out3d, RND));
		printf("       =   %14.7e\n", mpfr_get_d(out3, RND));
		printf("  eDi2 ( - %14.7e + %14.7e - %14.7e ) %14.7e\n",
		       mpfr_get_d(out1, RND), mpfr_get_d(out2, RND), mpfr_get_d(out3, RND),
		       mpfr_get_d(pre, RND)); printf("       =   %14.7e * %14.7e\n",
		       mpfr_get_d(tmp.w, RND), mpfr_get_d(pre, RND));
		printf("       =   %14.7e\n", mpfr_get_d(eDi2, RND));
		printf("  eDi  =   %14.7e\n", mpfr_get_d(eDi, RND));
		printf("  vDi  =   %14.7e\n", mpfr_get_d(res, RND));
		       
	}

	clear_scratch(&tmp);
	clear_incr3(&incr3a);
	clear_incr3(&incr3b);
	clear_converge(&sum3a);
	clear_converge(&sum3b);
	mpfr_clear(eDi);
	mpfr_clear(eDi2);
	mpfr_clear(out1);
	mpfr_clear(out2);
	mpfr_clear(out3);
	mpfr_clear(out3a);
	mpfr_clear(out3b);
	mpfr_clear(out3c);
	mpfr_clear(out3d);
	mpfr_clear(pre);
	mpfr_clear(tot);
	mpfr_clear(ksq);
	mpfr_clear(term);
}

/***
    var_Dlogistic_inner:  Calculate the inner integral (second moment) of the
                          spacing variance of the i'th point of n drawn from a
                          logistic distribution with scale/stdev s.  Evaluate
                          at the point z.
    args:                 res - inner integral
                          z - point/position
                          i - index of drawn sample
                          n - number of drawn samples
                          s - scale/stdev of distribution
                          trace - true to information about calculation
    modifies:  res
***/
static void var_Dlogistic_inner(mpfr_t res, double z, int i, int n, double s,
                                int trace) {
	mpfr_t term1;                         /* term without sum */
	mpfr_t term2;                         /* summed term with logarithm */
	mpfr_t term3;                         /* double summed term */
	mpfr_t tot;                           /* intermediate sum over l */
	mpfr_t u;                             /* intermediate values */
	mpfr_t v;
	mpfr_t w;
	mpfr_t x;
	mpfr_t y;
	int nu;                               /* n-i+2 */
	int k, l;

	mpfr_init2(term1, NBIT);
	mpfr_init2(term2, NBIT);
	mpfr_init2(term3, NBIT);
	mpfr_init2(tot, NBIT);
	mpfr_init2(u, NBIT);
	mpfr_init2(v, NBIT);
	mpfr_init2(w, NBIT);
	mpfr_init2(x, NBIT);
	mpfr_init2(y, NBIT);

	nu = n - i + 2;

	if (TRACE_GEN & trace) {
		printf("double sum\n");
		printf("  %3s   %12s * %12s = %12s   %12s   %3s\n",
		       "k", "exp/ni1", "lsum", "term", "cum", "ni1");
	}
	mpfr_set_zero(term1, 0);
	for (k=1; k<=(nu-3); k++) {
		mpfr_set_zero(tot, 0);
		for (l=1; l<=(nu-2-k); l++) {
			/* x = e^(l*z) */
			mpfr_set_d(u, z, RND);
			mpfr_mul_si(v, u, l, RND);
			mpfr_exp(x, v, RND);
			/* y = (1+e^-z)^-(nu-1-k-l) */
			mpfr_set_d(u, -z, RND);
			mpfr_exp(v, u, RND);
			mpfr_add_d(w, v, 1.0, RND);
			mpfr_pow_si(y, w, -(nu-1-k-l), RND);
			/* term3 = x * y / (nu-1-k-l), add into tot */
			mpfr_mul(v, x, y, RND);
			mpfr_div_si(w, v, nu-1-k-l, RND);
			mpfr_add(tot, tot, w, RND);
		}
		/* v = exp(k*z) / (nu-1-k), x = v * total l, add into term1 */
		mpfr_set_d(u, z, RND);
		mpfr_mul_si(v, u, k, RND);
		mpfr_exp(w, v, RND);
		mpfr_div_si(v, w, nu-1-k, RND);
		mpfr_mul(x, tot, v, RND);
		mpfr_add(term1, term1, x, RND);
		if (TRACE_DTL & trace) {
			printf("  %3d   %12.5e * %12.5e = %12.5e   %12.5e   %3d\n",
			       k, mpfr_get_d(v, RND), mpfr_get_d(tot, RND),
			       mpfr_get_d(x, RND), mpfr_get_d(term1, RND), nu-1-k);
		}
	}

	if (TRACE_GEN & trace) {
		printf("single sum 1/(nu-1-k)\n");
		printf("  %3s   %12s * %12s * %12s = %12s\n",
		       "", "sum", "exp", "ln", "term");
	}
	mpfr_set_zero(tot, 0);
	for (k=1; k<=(nu-2); k++) {
		/* 1/(nu-1-k) and add into tot */
		mpfr_set_d(u, nu-1-k, RND);
		mpfr_si_div(v, 1, u, RND);
		mpfr_add(tot, tot, v, RND);
	}
	/* x = e^((nu-1)*z) */
	mpfr_set_d(u, z, RND);
	mpfr_mul_si(v, u, nu-1, RND);
	mpfr_exp(x, v, RND);
	/* y = log(1+exp(-z)) */
	mpfr_set_d(u, -z, RND);
	mpfr_exp(v, u, RND);
	mpfr_add_d(w, v, 1.0, RND);
	mpfr_log(y, w, RND);

	/* term2 = x*y*tot */
	mpfr_mul(u, x, y, RND);
	mpfr_mul(term2, u, tot, RND);
	if (TRACE_GEN & trace) {
		printf("  %3s   %12.5e * %12.5e * %12.5e = %12.5e   %3s\n",
		       " ", mpfr_get_d(tot, RND), mpfr_get_d(x, RND), mpfr_get_d(y, RND),
		       mpfr_get_d(term2, RND), " ");
	}

	if (TRACE_GEN & trace) {
		printf("dilog term\n");
		printf("  %3s   %12s * %12s = %12s   %12s\n",
		       "", "exp", "dilog", "term", " ");
	}
	/* x = e^((nu-1)*z) already set */
	/* y = dilog(-exp(-z)) */
	mpfr_set_d(u, -z, RND);
	mpfr_exp(v, u, RND);
	mpfr_mul_d(w, v, -1.0, RND);
	mpfr_li2(y, w, RND);
	mpfr_mul(term3, x, y, RND);
	if (TRACE_GEN & trace) {
		printf("  %3s   %12.5e * %12.5e = %12.5e   %12s   %3s\n",
		       " ", mpfr_get_d(x, RND), mpfr_get_d(y, RND), mpfr_get_d(term3, RND),
		       " ", " ");
	}

	/* res = 2 * (term1 - term2 - term3) / (nu-1) */
	mpfr_sub(u, term1, term2, RND);
	mpfr_sub(x, u, term3, RND);
	mpfr_set_si(u, 2, RND);
	mpfr_div_si(v, u, nu-1, RND);
	mpfr_mul(res, x, v, RND);
	
	if (TRACE_GEN & trace) {
		printf("result\n");
		printf("  %3s ( %12.5e - %12.5e - %12.5e )* %6.4f  =  %12.5e\n",
		       "res", mpfr_get_d(term1, RND), mpfr_get_d(term2, RND),
		       mpfr_get_d(term3, RND), mpfr_get_d(v, RND), mpfr_get_d(res, RND));
	}

	mpfr_clear(u);
	mpfr_clear(v);
	mpfr_clear(w);
	mpfr_clear(x);
	mpfr_clear(y);
	mpfr_clear(term1);
	mpfr_clear(term2);
	mpfr_clear(term3);
}

/***
    var_Dlogistic_out3a:  Calculate the |z| < 1 dilog integral out3a
                          incrementally at series index k and spacing index i,
                          and evaluate it at t=1 and 0.
    args:                 res - dilog integral for k, i
                          prev - equation values at previous k
                          tmp - intermediate values/constants
                          k - dilog series index (>= 1)
                          i - spacing index (>= 2)
                          trace - true to print information about series
    modifies:  res, prev, tmp
***/
static void var_Dlogistic_out3a(mpfr_t res, struct increq *prev,
                                struct scratch *tmp, int k, int i, int trace) {

	if ((i - 1) == k) {
		/* Constant is the 1/j 2^j sum - ln 2 */
		sum_lnconst(prev->c, i, tmp);
	} else if (1 == k) {
		/* Constant is (2^(i-1) - 1) / (i-1)! */
		mpfr_set_si_2exp(tmp->u, 1, i-1, RND);
		mpfr_sub(tmp->v, tmp->u, tmp->one, RND);
		mpfr_div(prev->c, tmp->v, tmp->i1fact, RND);
	}
	
	if ((i - 1) == k) {
		/* Initialize scla = (-1)^(i-1) k! / (k+1-i)! = (-1)^(i-1) (i-1)! */
		mpfr_set(prev->scl, tmp->i1fact, RND);
		if ((i - 1) % 2) {
			mpfr_neg(prev->scl, prev->scl, RND);
		}
	} else if (1 == k) {
		/* Initialize scl = k! (i-k-2)! / 2^(i-1) = - (i-3)! / 2^(i-1) */
		/* Note i > 2 or we would match the first branch. */
		mpfr_set_si(tmp->u, i-2, RND);
		mpfr_gamma(tmp->v, tmp->u, RND);
		mpfr_div_2si(prev->scl, tmp->v, i-1, RND);
	} else {
		/* Scale scl = -scl * -k / (k+1-i) for either series */
		mpfr_mul_si(tmp->u, prev->scl, -k, RND);
		mpfr_div_si(prev->scl, tmp->u, k+1-i, RND);
	}

	if (i == k) {
		/* Initialize sum = (-1)^i / i! 2^(i-1) */
		mpfr_mul_2si(tmp->u, tmp->ifact, i-1, RND);
		if (i % 2) {
			mpfr_neg(tmp->u, tmp->u, RND);
		}
		mpfr_div(prev->sum, tmp->one, tmp->u, RND);
	} else if (k == (i-1)) {
		/* Sum skipped for this case, also for i==2 where (1-i)! is nonsense. */
		mpfr_set_zero(prev->sum, 0);
	} else if (1 == k) {
		/* Initialize sum = 1 / (i-2)! */
		mpfr_div_si(tmp->u, tmp->i1fact, i-1, RND);
		mpfr_div(prev->sum, tmp->one, tmp->u, RND);
	} else if (k < (i-1)) {
		/* Increment sum = sum + 1/k! (i-k-1)! */
		mpfr_set_si(tmp->u, k+1, RND);
		mpfr_gamma(tmp->v, tmp->u, RND);
		mpfr_set_si(tmp->u, i-k, RND);
		mpfr_gamma(tmp->w, tmp->u, RND);
		mpfr_mul(tmp->u, tmp->v, tmp->w, RND);
		mpfr_div(tmp->v, tmp->one, tmp->u, RND);
		mpfr_add(prev->sum, prev->sum, tmp->v, RND);
	} else {
		/* Increment sum = sum + ( (-1)^k (k-1) ! / 2^(i-1) k! ) */
		tmp->fnum[0] = k - i;
		tmp->fdenom[0] = k;
		simplify_factfrac(tmp->num, tmp->denom, tmp->fnum, 1, tmp->fdenom, 1);
		mpfr_div(tmp->u, tmp->num, tmp->denom, RND);
		mpfr_div_2si(tmp->v, tmp->u, i-1, RND);
		if (k % 2) {
			mpfr_sub(prev->sum, prev->sum, tmp->v, RND);
		} else {
			mpfr_add(prev->sum, prev->sum, tmp->v, RND);
		}
	}

	if (k < (i-1)) {
		/* out3a = scl * (c - sum) */
		mpfr_sub(tmp->u, prev->c, prev->sum, RND);
		mpfr_mul(res, prev->scl, tmp->u, RND);
		if (TRACE_DTL & trace) {
			printf("%4d al %11.4e *( %11.4e - %11.4e )= %13.6e\n",
			       k, mpfr_get_d(prev->scl, RND), mpfr_get_d(prev->c, RND),
			       mpfr_get_d(prev->sum, RND), mpfr_get_d(res, RND));
		}
	} else {
		/* out3a = scl * (sum + (-1)^i c) */
		if (i % 2) {
			mpfr_sub(tmp->u, prev->sum, prev->c, RND);
		} else {
			mpfr_add(tmp->u, prev->sum, prev->c, RND);
		}
		mpfr_mul(res, prev->scl, tmp->u, RND);
		if (TRACE_DTL & trace) {
			printf("%4d ah %11.4e *( %11.4e %c %11.4e )= %13.6e\n",
			       k, mpfr_get_d(prev->scl, RND), mpfr_get_d(prev->sum, RND),
			       (i % 2) ? '-' : '+', mpfr_get_d(prev->c, RND),
			       mpfr_get_d(res, RND));
		}
	}
}

/***
    var_Dlogistic_out3b:  Calculate the |z| < 1 dilog integral out3a
                          incrementally at series index k and spacing index i,
                          and evaluate it at t=1 and 0.
    args:                 res - dilog integral for k, i
                          prev - equation values at previous k
                          tmp - intermediate values/constants
                          k - dilog series index (>= 1)
                          i - spacing index (>= 2)
                          trace - true to print information about series
    modifies:  res, prev, tmp
***/
static void var_Dlogistic_out3b(mpfr_t res, struct increq *prev,
                                struct scratch *tmp, int k, int i, int trace) {
	
	/* Constant is the 1/j 2^j sum - ln 2 */
	if (1 == k) {
		sum_lnconst(prev->c, i, tmp);
	}

	if (1 == k) {
		/* Initialize scl = -(i+k-2)! / (k-1)!, ie. -(i-1)! */
		mpfr_set(prev->scl, tmp->i1fact, RND);
		mpfr_neg(prev->scl, prev->scl, RND);
	} else {
		/* Scale scl = -scl * (i+k-2) / (k-1) */
		mpfr_mul_si(tmp->u, prev->scl, -(i+k-2), RND);
		mpfr_div_si(prev->scl, tmp->u, k-1, RND);
	}

	if (2 == k) {
		/* Initialize sum = 1 / i! 2^(i-1) */
		mpfr_mul_2si(tmp->u, tmp->ifact, i-1, RND);
		mpfr_div(prev->sum, tmp->one, tmp->u, RND);
	} else if (2 < k) {
		/* Increment sum = sum + (-1)^k (k-2)! / 2^(i-1) * (i+k-2)! */
		tmp->fnum[0] = k - 2;
		tmp->fdenom[0] = i + k - 2;
		simplify_factfrac(tmp->num, tmp->denom, tmp->fnum, 1, tmp->fdenom, 1);
		mpfr_div(tmp->u, tmp->num, tmp->denom, RND);
		mpfr_div_2si(tmp->v, tmp->u, i-1, RND);
		if (k % 2) {
			mpfr_sub(prev->sum, prev->sum, tmp->v, RND);
		} else {
			mpfr_add(prev->sum, prev->sum, tmp->v, RND);
		}
	} else {
		mpfr_set_zero(prev->sum, 0);
	}

	/* out3b = scl * (sum + c) */
	mpfr_add(tmp->u, prev->sum, prev->c, RND);
	mpfr_mul(res, prev->scl, tmp->u, RND);
	if (TRACE_DTL & trace) {
		printf("%4d  b %11.4e *( %11.4e + %11.4e )= %13.6e\n",
		       k, mpfr_get_d(prev->scl, RND), mpfr_get_d(prev->sum, RND),
		       mpfr_get_d(prev->c, RND), mpfr_get_d(res, RND));
	}
}

/***
    sum_lnconst:  Calculate the 1/(j*(2^j)) - ln 2 sum from 1 to spacing index 
                  i-1, which is equivalent to 1/(j*(2^j)) from i to infinity.
                  Scale by (i-1)!.
    args:         res - sum
                  i - spacing index (i >= 2)
                  scratch - intermediate values/constants
    modifies:  res
***/
static void sum_lnconst(mpfr_t res, int i, struct scratch *tmp) {
	int j;

	mpfr_set_zero(res, 0);
	for (j=1; j<=(i-1); j++) {
		mpfr_set_si_2exp(tmp->u, j, j, RND);
		mpfr_div(tmp->v, tmp->one, tmp->u, RND);
		mpfr_add(res, res, tmp->v, RND);
	}
	mpfr_const_log2(tmp->u, RND);
	mpfr_sub(res, res, tmp->u, RND);

	mpfr_div(res, res, tmp->i1fact, RND);
}

/***
    accelerate:  Cache the value (assumed added sequentially) and calculate 
                 the accelerated alternating sum.  Do nothing if the cache
                 is full.
    args:        conv - cache of alternating series values and predicted sum
                 term - next value in alternating series
    modifies:  conv
***/
static void accelerate(struct converge *conv, mpfr_t term, int trace) {
	int n;                                /* number of terms in series */
	int k;                                /* algorithm index */

	if (NCONV < conv->nconv) {
		return;
	}

	mpfr_set(conv->val[conv->nconv], term, RND);
	conv->nconv += 1;
	n = conv->nconv;

	/* Initialize for k=0.  c = d-1, b = 2*n*n. */
	mpfr_sqrt_ui(conv->u, 8, RND);
	mpfr_add_si(conv->u, conv->u, 3, RND);
	mpfr_pow_si(conv->u, conv->u, n, RND);
	mpfr_si_div(conv->d, 1, conv->u, RND);
	mpfr_add(conv->d, conv->d, conv->u, RND);
	mpfr_div_si(conv->d, conv->d, 2, RND);
	mpfr_set(conv->c, conv->d, RND);
	mpfr_sub_si(conv->c, conv->c, 1, RND);
	mpfr_set_si(conv->b, n, RND);
	mpfr_sqr(conv->b, conv->b, RND);
	mpfr_mul_si(conv->b, conv->b, 2, RND);
	mpfr_set_d(conv->s, 0.0, RND);
	if (trace && (n < 10)) {
		printf("bcd  %19.12e  %19.12e     %19.12e\n", mpfr_get_d(conv->b,RND),
		       mpfr_get_d(conv->c,RND), mpfr_get_d(conv->d,RND));
	}

	for (k=1; k<n; k++) {
		mpfr_sub(conv->c, conv->b, conv->c, RND);
		mpfr_mul(conv->u, conv->c, conv->val[k-1], RND);
		/* Algorithm series factors out (-1)^k from the a_k terms. */
		if (k % 2) {
			mpfr_sub(conv->s, conv->s, conv->u, RND);
		} else {
			mpfr_add(conv->s, conv->s, conv->u, RND);
		}
		mpfr_mul_si(conv->b, conv->b, 2 * (k+n) * (k-n), RND);
		mpfr_div_si(conv->b, conv->b, ((2*k)+1) * (k+1), RND);
		if (trace && (n < 10)) {
			printf("  %2d %19.12e  %19.12e     %19.12e\n", k,
			       mpfr_get_d(conv->b,RND), mpfr_get_d(conv->c,RND),
			       mpfr_get_d(conv->s,RND));
		}
	}
	
	mpfr_div(conv->pred[n-1], conv->s, conv->d, RND);
	if (trace) {
		printf("%-3d  %19.12e  %19.12e", conv->nconv, mpfr_get_d(term,RND),
		       mpfr_get_d(conv->pred[n-1],RND));
		if (1 < n) {
			mpfr_sub(conv->u, conv->pred[n-1], conv->pred[n-2], RND);
			mpfr_abs(conv->u, conv->u, RND);
			printf("     %19.12e\n", mpfr_get_d(conv->u,RND));
		} else {
			printf("\n");
		}
	}
}

/***
    accelerate_R:  Accelerate the convergence of an alternating sum, using
                   the R version of the Cohen/Villegas/Zagier algorithm.
    args:          conv - cache of alternating series values and predicted sum
                   term - next value in alternating series
    modifies:  conv
***/
static void accelerate_R(struct converge *conv, mpfr_t term, int trace) {
	int n;                                /* abbreviation number data points */
	int k;

	if (NCONV < conv->nconv) {
		return;
	}

	/* This seems to be a re-working of the Cohen algorithm 1, counting k
	   down rather than up (note b scaling is reciprocal of base version).
	*/

	/* val and pred are 0 based. */
	mpfr_set(conv->val[conv->nconv], term, RND);
	conv->nconv += 1;
	/* This makes the equations a bit nicer. */
	n = conv->nconv;
	
	mpfr_set_si_2exp(conv->b, 1, (2*n)-1, RND);
	mpfr_set(conv->c, conv->b, RND);
	mpfr_set_zero(conv->s, 0);

	for (k=n-1; 0<k; k--) {
		mpfr_mul(conv->u, conv->c, conv->val[k-1], RND);
		mpfr_add(conv->s, conv->s, conv->u, RND);
		mpfr_set_si(conv->u, ((2*k)+1) * (k+1), RND);
		mpfr_div_si(conv->u, conv->u, 2 * (n-k) * (n+k), RND);
		mpfr_mul(conv->b, conv->b, conv->u, RND);
		mpfr_add(conv->c, conv->c, conv->b, RND);
	}

	mpfr_set_si(conv->u, 2*n*n, RND);
	mpfr_div(conv->b, conv->b, conv->u, RND);
	mpfr_add(conv->c, conv->c, conv->b, RND);
	mpfr_div(conv->pred[n-1], conv->s, conv->c, RND);
	if (trace) {
		printf("%-3d  %19.12e  %19.12e", conv->nconv, mpfr_get_d(term,RND),
		       mpfr_get_d(conv->pred[n-1],RND));
		if (1 < n) {
			mpfr_sub(conv->u, conv->pred[n-1], conv->pred[n-2], RND);
			mpfr_abs(conv->u, conv->u, RND);
			printf("     %19.12e\n", mpfr_get_d(conv->u,RND));
		} else {
			printf("\n");
		}
	}
}

/***
    simplify_frac:  Make the representation of a fraction consistent.
                    Move any sign of the numerator and denominator into the
                    integer and remove any common (gcd) factor from both.
    args:           sgn - sign flag for fraction
                    num - numerator
                    denom - denominator
    modifies:  sgn, num, denom
***/
static void simplify_frac(int *sgn, mpfr_t num, mpfr_t denom) {
	mpfr_t common;                        /* common factor to num, denom */

	mpfr_init2(common, NBIT);
	
	if (mpfr_sgn(num) < 0) {
		mpfr_neg(num, num, RND);
		*sgn = -*sgn;
	}
	if (mpfr_sgn(denom) < 0) {
		mpfr_neg(denom, denom, RND);
		*sgn = -*sgn;
	}

	mpfr_gcd(common, num, denom);
	if (0 <= mpfr_cmp_ui(common, 1)) {
		mpfr_div(num, num, common, RND);
		mpfr_div(denom, denom, common, RND);
	}

	mpfr_clear(common);
}

/***
    simplify_factfrac:  Simplify the fraction formed by the factorial of
                        each element of vnum divided by that of vdenom,
                        ie. (vnum[1]! * vnum[2]! * ...) /
                            (vdenom[1]! * vdenom[2]! * ...)
                        Leave as two separate high-precision numbers for the
                        rational result.
    args:               num, denom - numerator, denominator of result
                        vnum - numerator values to take factorial, product of
                        nnum - number of values in vnum
                        vdenom - denominator values for factorial, product
                        ndenom - number of values in vdenom
    modifies:  num, denom
***/
static void simplify_factfrac(mpfr_t num, mpfr_t denom, int *vnum, int nnum,
                              int *vdenom, int ndenom) {
	int *nvals;                           /* all numerator values */
	int *dvals;                           /* all denominator values */
	int nn;                               /* number of entries stored in nvals */
	int nd;                               /* number of entries stored in dvals */
	int nnvals;                           /* number of nnvals allocated */
	int ndvals;                           /* number of ndvals allcoted */
	int common;                           /* common (gcd) factor */
	int changed;                          /* mark if values changed */
	int i, j, k;

	nvals = NULL;
	dvals = NULL;

	nnvals = 0;
	for (i=0; i<nnum; i++) {
		if (1 < vnum[i]) {
			/* We include vnum[i], but exclude the 1 so no need to add 1. */
			nnvals += vnum[i];
		}
	}
	if (0 == nnvals) {
		nn = 0;
	} else {
		nvals = calloc(nnvals+1, sizeof(*nvals));
		if (NULL == nvals) {
			fprintf(stderr, "couldn't allocate internal numerator list\n");
			goto cleanup;
		}

		nn = 0;
		for (i=0; i<nnum; i++) {
			for (j=2; j<=vnum[i]; j++,nn++) {
				nvals[nn] = j;
			}
		}
	}

	ndvals = 0;
	for (i=0; i<ndenom; i++) {
		if (1 < vdenom[i]) {
			ndvals += vdenom[i];
		}
	}
	if (0 == ndvals) {
		nd = 0;
	} else {
		dvals = calloc(ndvals+1, sizeof(*dvals));
		if (NULL == dvals) {
			fprintf(stderr, "couldn't allocate internal denominator list\n");
			goto cleanup;
		}
		
		nd = 0;
		for (i=0; i<ndenom; i++) {
			for (j=2; j<=vdenom[i]; j++,nd++) {
				dvals[nd] = j;
			}
		}
	}
	
	if ((0 == nn) || (0 == nd)) {
		goto cleanup;
	}

	/* First remove any factors that are in both product lists. */
	qsort(nvals, nn, sizeof(*nvals), sortfn_int);
	qsort(dvals, nd, sizeof(*dvals), sortfn_int);

	i = 0;
	j = 0;
	while ((i < nn) && (j < nd)) {
		if (nvals[i] == dvals[j]) {
			nvals[i] = 0;
			dvals[j] = 0;
			i += 1;
			j += 1;
		} else if (nvals[i] < dvals[j]) {
			i += 1;
		} else {
			j += 1;
		}
	}
	remove_nonpos(nvals, &nn);
	remove_nonpos(dvals, &nd);
	if ((0 == nn) || (0 == nd)) {
		goto cleanup;
	}

	/* Second remove direct multiples.  Continue until all exhausted. */
	changed = 0x11;
	while (changed && (0 < nn) && (0 < nd)) {
		changed = 0;
		for (i=0; i<nn; i++) {
			for (j=0; j<nd; j++) {
				if (0 == (dvals[j] % nvals[i])) {
					dvals[j] /= nvals[i];
					j = nd + 1;
					nvals[i] = 0;
					changed |= 0x01;
				}
			}
		}
		if (changed & 0x01) {
			remove_nonpos(nvals, &nn);
			qsort(dvals, nd, sizeof(*dvals), sortfn_int);
		}

		for (j=0; j<nd; j++) {
			for (i=0; i<nn; i++) {
				if (0 == (nvals[i] % dvals[j])) {
					nvals[i] /= dvals[j];
					i = nn + 1;
					dvals[j] = 0;
					changed |= 0x10;
				}
			}
		}
		if (changed & 0x10) {
			remove_nonpos(dvals, &nd);
			qsort(nvals, nn, sizeof(*nvals), sortfn_int);
		}
	}
	if ((0 == nn) || (0 == nd)) {
		goto cleanup;
	}

	/* Third remove partial multiples.  Single pass of all num-denom pairs. */
	for (i=0; i<nn; i++) {
		for (j=0; j<nd; j++) {
			common = gcd(nvals[i], dvals[j]);
			if (1 < common) {
				dvals[j] /= common;
				nvals[i] /= common;
			}
		}
	}
	
cleanup:
	if (0 == nn) {
		mpfr_set_ui(num, 1, RND);
	} else {
		mpfr_set_ui(num, nvals[0], RND);
		for (i=1; i<nn; i++) {
			mpfr_mul_ui(num, num, nvals[i], RND);
		}
	}
	if (0 == nd) {
		mpfr_set_ui(denom, 1, RND);
	} else {
		mpfr_set_ui(denom, dvals[0], RND);
		for (i=1; i<nd; i++) {
			mpfr_mul_ui(denom, denom, dvals[i], RND);
		}
	}
	
	free(nvals);
	free(dvals);
}

/***
    sortfn_int:  Comparison function between two integers, increasing order.
    args:        val1, val2 - values to compare
    returns:   < 0 if val1 before val2, > 0 if after, 0 if same
***/
static int sortfn_int(const void *val1, const void *val2) {
	int *ival1;                           /* argument converted to integer */
	int *ival2;

	ival1 = (int *) val1;
	ival2 = (int *) val2;
	return (*ival1 - *ival2);
}

/***
    remove_nonpos:  Compact the vals array by moving all valid (positive)
                    values to the front.  Will adjust the number of values
                    in the array.  The order of elements is preserved.
    args:           vals - array of values to compact
                    nval - number of values in array
    modifies:  vals, nval
***/
static void remove_nonpos(int *vals, int *nval) {
	int i, j;

	for (i=0,j=0; i<*nval; i++) {
		if (i != j) {
			vals[j] = vals[i];
		}
		if (0 < vals[i]) {
			j += 1;
		}
	}
	*nval = j;
}

/***
    gcd:  Calculate the greatest common divisor of two integers.
    args:  a, b - integers to find gcd of
    returns:   gcd(a,b) (>= 1)
***/
static int gcd(int a, int b) {
	double u, v;                          /* reducing versions of arguments */
	double q;                             /* integer quotient of u/v */
	double t;                             /* remainder of u */

	if ((0 == a) || (0 == b)) {
		return 0;
	}

	/* This is the algorithm used in pracma's gcd function. */
	u = abs(a);
	v = abs(b);
	while (v != 0) {
		q = floor(u/v);
		t = u - (v * q);
		u = v;
		v = t;
	}
	lround(u);
}

/***
    mpfr_gcd:  High-precision library version of gcd, calculating the
               greatest common divisor of two integers.
    args:      rop - gcd of other arguments
               op1, op2 - integers to find gcd of
    modifies:  common
***/
static void mpfr_gcd(mpfr_t rop, mpfr_t op1, mpfr_t op2) {
	mpfr_t u;                             /* reducing version of inputs */
	mpfr_t v;
	mpfr_t q;                             /* integer quotient of u/v */
	mpfr_t t;                             /* remainder of u */
	mpfr_t w;                             /* temporary variable */

	mpfr_init2(u, NBIT);
	mpfr_init2(v, NBIT);
	mpfr_init2(q, NBIT);
	mpfr_init2(t, NBIT);
	mpfr_init2(w, NBIT);

	if (mpfr_zero_p(op1) || mpfr_zero_p(op2)) {
		mpfr_set_ui(rop, 0, RND);
	} else {
		mpfr_abs(u, op1, RND);
		mpfr_abs(v, op2, RND);
		while (!mpfr_zero_p(v)) {
			mpfr_div(w, u, v, RND);
			mpfr_floor(q, w);
			mpfr_mul(w, v, q, RND);
			mpfr_sub(t, u, w, RND);
			mpfr_round(t, t);
			mpfr_set(u, v, RND);
			mpfr_set(v, t, RND);
		}
	}
	mpfr_round(rop, u);

	mpfr_clear(u);
	mpfr_clear(v);
	mpfr_clear(q);
	mpfr_clear(t);
	mpfr_clear(w);
}

/***
    init_scratch:  Initialize the mpfr values in the scratch data.  Set up
                   any constants.
    args:          tmp - temporary storage or constant values
                   i - spacing index (>= 2)
    modifies:  tmp
***/
static void init_scratch(struct scratch *tmp, int i) {
	
	mpfr_init2(tmp->u, NBIT);
	mpfr_init2(tmp->v, NBIT);
	mpfr_init2(tmp->w, NBIT);
	mpfr_init2(tmp->num, NBIT);
	mpfr_init2(tmp->denom, NBIT);
	mpfr_init2(tmp->one, NBIT);
	mpfr_init2(tmp->i1fact, NBIT);
	mpfr_init2(tmp->ifact, NBIT);
	/* Don't initialize the fnum/fdenom arrays - not needed. */
	
	mpfr_set_si(tmp->one, 1, RND);

	/* (i-1)! = Gamma(i) */
	mpfr_set_si(tmp->u, i, RND);
	mpfr_gamma(tmp->i1fact, tmp->u, RND);

	mpfr_mul_si(tmp->ifact, tmp->i1fact, i, RND);
}

/***
    init_incr3:  Initialize the mpfr values in the incremental equation for
                 the dilog integrals.
    args:        eqn - incremental equation for dilog integral component
    modifies:  eqn
***/
static void init_incr3(struct increq *eqn) {
	
	mpfr_init2(eqn->scl, NBIT);
	mpfr_init2(eqn->sum, NBIT);
	mpfr_init2(eqn->c, NBIT);
}

/***
    init_converge:  Initialize convergence tracking.
    args:           conv - cache of alternating series values
    modifies:  conv
***/
static void init_converge(struct converge *conv) {
	int i;

	for (i=0; i<NCONV; i++) {
		mpfr_init2(conv->val[i], NBIT);
		mpfr_init2(conv->pred[i], NBIT);
	}
	mpfr_init2(conv->b, NBIT);
	mpfr_init2(conv->c, NBIT);
	mpfr_init2(conv->d, NBIT);
	mpfr_init2(conv->s, NBIT);
	mpfr_init2(conv->u, NBIT);

	mpfr_set_d(conv->val[0], 0.0, NBIT);
	mpfr_set_d(conv->pred[0], 0.0, NBIT);
	conv->nconv = 0;
}

/***
    clear_scratch:  Finish using any mpfr values in the scratch data.  Do not
                    use after calling this (or, do an init_scratch).
    args:           tmp - temporary storage or constant values
    modifies:  tmp
***/
static void clear_scratch(struct scratch *tmp) {

	mpfr_clear(tmp->u);
	mpfr_clear(tmp->v);
	mpfr_clear(tmp->w);
	mpfr_clear(tmp->num);
	mpfr_clear(tmp->denom);
	mpfr_clear(tmp->one);
	mpfr_clear(tmp->i1fact);
	mpfr_clear(tmp->ifact);
}

/***
    clear_incr3:  Finish using any mpfr values in the incremental equation for
                  the dilog integrals.
    args:         eqn - incremental equation for dilog integral component
    modifies:  eqn
***/
static void clear_incr3(struct increq *eqn) {

	mpfr_clear(eqn->scl);
	mpfr_clear(eqn->sum);
	mpfr_clear(eqn->c);
}

/***
    clear_converge:  Finish using any mpfr values in the convergence cache.
                     Do not use after calling this (nconv set < 0 to mark).
    args:            conv - cache of alternating series values
    modifies:  conv
***/
static void clear_converge(struct converge *conv) {
	int i;

	for (i=0; i<NCONV; i++) {
		mpfr_clear(conv->val[i]);
		mpfr_clear(conv->pred[i]);
	}
	mpfr_clear(conv->b);
	mpfr_clear(conv->c);
	mpfr_clear(conv->d);
	mpfr_clear(conv->s);
	mpfr_clear(conv->u);
	conv->nconv = -1;
}



/**** Test functions ****/

/***
    BIST_gcd:  Test the gcd() function.
    returns:   true if all tests pass, 0 if any fail
***/
static int BIST_gcd(void) {
	int res;                              /* gcd result */
	int passed;                           /* overall pass/fail flag */

	passed = 1;
	printf("  starting gcd() tests\n");

	res = gcd(4, 12);
	if (4 != res) {
		printf("      gcd(4, 12) returned %d, not 4\n", res);
		passed = 0;
	}

	res = gcd(8, 12);
	if (4 != res) {
		printf("      gcd(8, 12) returned %d, not 4\n", res);
		passed = 0;
	}

	res = gcd(4, 1);
	if (1 != res) {
		printf("      gcd(4, 1) returned %d, not 1\n", res);
		passed = 0;
	}

	res = gcd(-105, 84);
	if (21 != res) {
		printf("      gcd(-105, 84) returned %d, not 21\n", res);
		passed = 0;
	}

	res = gcd(0, 9);
	if (0 != res) {
		printf("      gcd(0, 9) returned %d, not 0\n", res);
		passed = 0;
	}

	res = gcd(9, 0);
	if (0 != res) {
		printf("      gcd(9, 0) returned %d, not 0\n", res);
		passed = 0;
	}

	res = gcd(51, 37);
	if (1 != res) {
		printf("      gcd(51, 37) returned %d, not 1\n", res);
		passed = 0;
	}

	if (passed) {
		printf("    all tests passed\n");
	} else {
		printf("    some test FAILED\n");
	}
	return passed;
}

/***
    BIST_mpfr_gcd:  Test the mpfr_gcd() function.
    returns:   true if all tests pass, 0 if any fail
***/
static int BIST_mpfr_gcd(void) {
	mpfr_t val1;                          /* first argument to gcd */
	mpfr_t val2;                          /* second argument to gcd */
	mpfr_t res;                           /* common factor */
	int ires;                             /* result as an integer */
	int passed;                           /* overall pass/fail flag */

	mpfr_init2(val1, NBIT);
	mpfr_init2(val2, NBIT);
	mpfr_init2(res, NBIT);

	passed = 1;
	printf("  starting mpfr_gcd() tests\n");

	mpfr_set_si(val1, 4, RND);
	mpfr_set_si(val2, 12, RND);
	mpfr_gcd(res, val1, val2);
	ires = mpfr_get_si(res, RND);
	if (4 != ires) {
		printf("        mpfr_gcd(4, 12) returned %d, not 4\n", ires);
		passed = 0;
	}

	mpfr_set_si(val1, 8, RND);
	mpfr_set_si(val2, 12, RND);
	mpfr_gcd(res, val1, val2);
	ires = mpfr_get_si(res, RND);
	if (4 != ires) {
		printf("        mpfr_gcd(4, 8) returned %d, not 4\n", ires);
		passed = 0;
	}
	
	mpfr_set_si(val1, 4, RND);
	mpfr_set_si(val2, 1, RND);
	mpfr_gcd(res, val1, val2);
	ires = mpfr_get_si(res, RND);
	if (1 != ires) {
		printf("        mpfr_gcd(4, 1) returned %d, not 1\n", ires);
		passed = 0;
	}
	
	mpfr_set_si(val1, -105, RND);
	mpfr_set_si(val2, 84, RND);
	mpfr_gcd(res, val1, val2);
	ires = mpfr_get_si(res, RND);
	if (21 != ires) {
		printf("        mpfr_gcd(-105, 84) returned %d, not 21\n", ires);
		passed = 0;
	}
	
	mpfr_set_si(val1, 0, RND);
	mpfr_set_si(val2, 9, RND);
	mpfr_gcd(res, val1, val2);
	ires = mpfr_get_si(res, RND);
	if (0 != ires) {
		printf("        mpfr_gcd(0, 9) returned %d, not 0\n", ires);
		passed = 0;
	}
	
	mpfr_set_si(val1, 9, RND);
	mpfr_set_si(val2, 0, RND);
	mpfr_gcd(res, val1, val2);
	ires = mpfr_get_si(res, RND);
	if (0 != ires) {
		printf("        mpfr_gcd(9, 0) returned %d, not 0\n", ires);
		passed = 0;
	}
	
	mpfr_set_si(val1, 51, RND);
	mpfr_set_si(val2, 37, RND);
	mpfr_gcd(res, val1, val2);
	ires = mpfr_get_si(res, RND);
	if (1 != ires) {
		printf("        mpfr_gcd(51, 37) returned %d, not 1\n", ires);
		passed = 0;
	}

	mpfr_clear(val1);
	mpfr_clear(val2);
	mpfr_clear(res);
	
	if (passed) {
		printf("    all tests passed\n");
	} else {
		printf("    some test FAILED\n");
	}
	return passed;
}

/***
    BIST_remove_nonpos:  Test the remove_nonpos() function.
    returns:   true if all tests pass, 0 if any fail
***/
static int BIST_remove_nonpos(void) {
	int vals1[] = { 1, 2, 4, 3, 6, 2 };   /* test array 1 */
	int res1[] = { 1, 2, 4, 3, 6, 2 };    /* expected result test 1 */
	int vals2[] = { 1, 0, 4, 3, 0, 6, 2}; /* test array 2 */
	int res2[] = { 1, 4, 3, 6, 2 };       /* expected result test 2 */
	int vals3[] = { 0, 2, 6, 0, 3, 0 };   /* test array 3 */
	int res3[] = { 2, 6, 3 };             /* expected result test 3 */
	int nval;                             /* number of values in list */
	int pass;                             /* single test pass */
	int passed;                           /* overall pass/fail flag */
	int i;

	passed = 1;
	printf("  starting remove_nonpos() tests\n");

	nval = LEN(vals1);
	remove_nonpos(vals1, &nval);
	if (!BIST_match_nonpos(1, vals1, nval, res1, LEN(res1))) {
		passed = 0;
	}

	nval = sizeof(vals2) / sizeof(*vals2);
	remove_nonpos(vals2, &nval);
	if (!BIST_match_nonpos(2, vals2, nval, res2, LEN(res2))) {
		passed = 0;
	}

	nval = sizeof(vals3) / sizeof(*vals3);
	remove_nonpos(vals3, &nval);
	if (!BIST_match_nonpos(3, vals3, nval, res3, LEN(res3))) {
		passed = 0;
	}
	
	if (passed) {
		printf("    all tests passed\n");
	} else {
		printf("    some test FAILED\n");
	}
	return passed;
}

/***
    BIST_match_nonpos:  Compare the returned array to the expected.
                        Prints the mismatch if it exists.
    args:               tstID - identifier for error message
                        vals - values to compare to expected
                        nval - number of items in vals
                        exp - expected values to find
                        nexp - number of items in exp
    returns:   true if arrays match, 0 if not
***/
static int
BIST_match_nonpos(int tstID, int *vals, int nval, int *exp, int nexp) {
	int pass;                             /* pass/fail flag */
	int i;

	if (nval != nexp) {
		pass = 0;
	} else {
		pass = 1;
		for (i=0; i<nval; i++) {
			if (vals[i] != exp[i]) {
				pass = 0;
			}
		}
	}

	if (!pass) {
		printf("    test %d    (%2d):", tstID, nexp);
		for (i=0; i<nexp; i++) {
			printf("  %2d", exp[i]);
		}
		printf("\n");
		printf("      but got (%2d):", nval);
		for (i=0; i<nval; i++) {
			printf("  %2d", vals[i]);
		}
		printf("\n");
	}

	return pass;
}

/***
    BIST_simplify_factfrac:  Test the simplify_factfrac() function.
    returns:   true if all tests pass, 0 if any fail.
***/
static int BIST_simplify_factfrac(void) {
	mpfr_t fnum;                          /* numerator of result */
	mpfr_t fdenom;                        /* denominator of result */
	int num1[] = { 38, 7 };               /* test 1 inputs */
	int denom1[] = { 20, 25 };
	int expnum1 = 698523;                 /* test 1 results */
	int expdenom1 = 10;
	int num2[] = { 10, 12 };              /* test 2 input */
	int denom2[] = { 2, 15, 6 };
	int expnum2 = 12;                     /* test 2 results */
	int expdenom2 = 13;
	int num3[] = { 4, 5, 6, 7 };          /* test 3 input */
	int denom3[] = { 2, 4 };
	int expnum3 = 217728000;              /* test 3 results */
	int expdenom3 = 1;
	int num4[] = { -10, 0, 1 };           /* test 4 input */
	int denom4[] = { 1 };
	int expnum4 = 1;                      /* test 4 results */
	int expdenom4 = 1;
	int num5[] = { 6 };                   /* test 5 input */
	int denom5[] = { 7 };
	int expnum5 = 1;                      /* test 5 results */
	int expdenom5 = 7;
	int passed;                           /* overall pass/fail flag */

	mpfr_init2(fnum, NBIT);
	mpfr_init2(fdenom, NBIT);

	passed = 1;
	printf("  starting simplify_factfrac() tests\n");

	simplify_factfrac(fnum, fdenom, num1, LEN(num1), denom1, LEN(denom1));
	if ((expnum1 != mpfr_get_si(fnum, RND)) ||
	    (expdenom1 != mpfr_get_si(fdenom, RND))) {
		printf("    test 1: expect %d / %d, got %d / %d\n",
		       expnum1, expdenom1,
		       mpfr_get_si(fnum, RND), mpfr_get_si(fdenom, RND));
		passed = 0;
	}

	simplify_factfrac(fnum, fdenom, num2, LEN(num2), denom2, LEN(denom2));
	if ((expnum2 != mpfr_get_si(fnum, RND)) ||
	    (expdenom2 != mpfr_get_si(fdenom, RND))) {
		printf("    test 2: expect %d / %d, got %d / %d\n",
		       expnum2, expdenom2,
		       mpfr_get_si(fnum, RND), mpfr_get_si(fdenom, RND));
		passed = 0;
	}

	simplify_factfrac(fnum, fdenom, num3, LEN(num3), denom3, LEN(denom3));
	if ((expnum3 != mpfr_get_si(fnum, RND)) ||
	    (expdenom3 != mpfr_get_si(fdenom, RND))) {
		printf("    test 3: expect %d / %d, got %d / %d\n",
		       expnum3, expdenom3,
		       mpfr_get_si(fnum, RND), mpfr_get_si(fdenom, RND));
		passed = 0;
	}

	simplify_factfrac(fnum, fdenom, num4, LEN(num4), denom4, LEN(denom4));
	if ((expnum4 != mpfr_get_si(fnum, RND)) ||
	    (expdenom4 != mpfr_get_si(fdenom, RND))) {
		printf("    test 4: expect %d / %d, got %d / %d\n",
		       expnum4, expdenom4,
		       mpfr_get_si(fnum, RND), mpfr_get_si(fdenom, RND));
		passed = 0;
	}

	simplify_factfrac(fnum, fdenom, num5, LEN(num5), denom5, LEN(denom5));
	if ((expnum5 != mpfr_get_si(fnum, RND)) ||
	    (expdenom5 != mpfr_get_si(fdenom, RND))) {
		printf("    test 5: expect %d / %d, got %d / %d\n",
		       expnum5, expdenom5,
		       mpfr_get_si(fnum, RND), mpfr_get_si(fdenom, RND));
		passed = 0;
	}

	mpfr_clear(fnum);
	mpfr_clear(fdenom);
	
	if (passed) {
		printf("    all tests passed\n");
	} else {
		printf("    some test FAILED\n");
	}
	return passed;
}

/***
    BIST_init_scratch:  Test the constants set in the scratch data for
                        different spacing indices.
    returns:   true if all tests pass, 0 if any fail
***/
static int BIST_init_scratch(void) {
	struct scratch tmp;                   /* intermediate values/constants */
	double dval;                          /* value from structure */
	int passed;                           /* overall pass/fail flag */

	passed = 1;
	printf("  starting init_scratch() tests\n");

	init_scratch(&tmp, 5);
	dval = mpfr_get_d(tmp.one, RND);
	if (1.0 != dval) {
		printf("      one %13.6e is not 1 (i=5)\n", dval);
		passed = 0;
	}
	dval = mpfr_get_d(tmp.i1fact, RND);
	if (24.0 != dval) {
		printf("      (i-1)! %13.6e is not 24 (i=5)\n", dval);
		passed = 0;
	}
	dval = mpfr_get_d(tmp.ifact, RND);
	if (120.0 != dval) {
		printf("      i! %13.6e is not 120 (i=5)\n", dval);
		passed = 0;
	}
	clear_scratch(&tmp);

	init_scratch(&tmp, 11);
	if (1.0 != mpfr_get_d(tmp.one, RND)) {
		printf("      one %13.6e is not 1 (i=11)\n", tmp.one);
		passed = 0;
	}
	dval = mpfr_get_d(tmp.i1fact, RND);
	if (3628800.0 != dval) {
		printf("      (i-1)! %13.6e is not 3,628,880 (i=11)\n", dval);
		passed = 0;
	}
	dval = mpfr_get_d(tmp.ifact, RND);
	if (39916800 != dval) {
		printf("      i! %13.6e is not 39,916,800 (i=11)\n", dval);
		passed = 0;
	}
	clear_scratch(&tmp);

	if (passed) {
		printf("    all tests passed\n");
	} else {
		printf("    some test FAILED\n");
	}
	return passed;
}



/**** Top Level ****/

/***
    usage:  Print help message and exit.  The message will be tterminated with
            a newline if one isn't provided.  Requires the command line
            be stored in a global array of strings.
    args:   msg - reason for usage, may be NULL, treated as printf fmt string
            ... - arguments for fmt
***/
static void usage(const char *msg, ...) {
	va_list args;                         /* extra arguments for printf */
	int i;

	if ((NULL != msg) && (0 < strlen(msg))) {
		printf("\nnote: ");
		va_start(args, msg);
		vprintf(msg, args);
		if ('\n' != msg[strlen(msg)-1]) {
			printf("\n");
		}
	}

	printf("\n");
	for (i=0; NULL != usage_cmdline[i]; i++) {
		printf("%6s   ", (0 == i) ? "Usage:" : "");
		printf("%s\n", usage_cmdline[i]);
	}
	printf("\n");	exit(1);
}

/***
    parse_commandline:  Handle the command line.  Exit through usage() if
                        needed.
    args:               argc - number of arguments
                        argv - the arguments
                        args - parsed command line
    modifies:  args
***/
static void parse_commandline(int argc, char **argv, struct cmdargs *args) {
	int nnum;                             /* number of numeric args parsed */
	int i;

	memset(args, 0, sizeof(*args));
	args->action = ACT_VDI;
	args->s = 1.0;
	nnum = 0;

	if (argc < 2) {
		usage(NULL);
	}
	if (has_helpopt(argc, argv)) {
		usage(NULL);
	}

	for (i=1; i<argc; i++) {
		if (match_arg(argv[i], "-bist", 2)) {
			args->action = ACT_BIST;
		} else if (match_arg(argv[i], "-cum", 2)) {
			args->cum = 1;
		} else if (match_arg(argv[i], "-detail", 2)) {
			args->trace |= TRACE_DTL;
		} else if (match_arg(argv[i], "-inner", 2)) {
			args->action = ACT_INNER;
		} else if (match_arg(argv[i], "-trace", 2)) {
			args->trace |= TRACE_GEN;
		} else {
			nnum += 1;
			switch (nnum) {
			case 1:
				if (!strcmp(argv[i], "-")) {
					continue;
				}
				args->i = atoi(argv[i]);
				break;
			case 2:
				if (!strcmp(argv[i], "-")) {
					continue;
				}
				args->n = atoi(argv[i]);
				break;
			case 3:
				if (!strcmp(argv[i], "-")) {
					continue;
				}
				args->s = atof(argv[i]);
				break;
			case 4:
				if (!strcmp(argv[i], "-")) {
					continue;
				}
				args->z = atof(argv[i]);
				break;
			default:
				usage("unrecognized argument '%s'", argv[i]);
			}
		}
	}

	if (ACT_BIST == args->action) {
		return;
	}

	if (0 == nnum) {
		usage("must supply n");
	}
	if (1 == nnum) {
		args->n = args->i;
		args->i = 0;
	}
	
	if (args->n < 2) {
		usage("n must be >= 2");
	}
	if (args->s <= 0.0) {
		usage("s must be positive");
	}
}

/***
    has_helpopt:  Test if the first command line option is "-h" or "-help"
                  (normally you would call usage(NULL) to print the command
                  line).
    args:         argc - number of command line arguments
                  argv - the arguments
    returns:   true if the first option is "-h" or "-help", 0 if not (no error)
***/
static int has_helpopt(int argc, char **argv) {
	int err;
	if (1 < argc) {
		return !strcmp(argv[1], "-h") || !strcmp(argv[1], "-help");
	}
	return 0;
}

/***
   match_arg:  Test if two strings are equal, matching all of str1 or at
               least nchar characters.
   args:       str1, str2 - strings to compare
               nchar - minimum length of string match
   returns:  true if strings the same, 0 if not
***/
static int match_arg(const char *str1, const char *str2, unsigned int nchar) {
	unsigned int len1;                    /* length of str1 */

	len1 = strlen(str1);
	if (len1 < nchar) {
		return 0;
	}
	if (strlen(str2) < len1) {
		return 0;
	}
	return !strncmp(str1, str2, len1);
}


int main(int argc, char **argv) {
	struct cmdargs args;                  /* command line arguments */
	struct increq incr3a;                 /* incremental out3a values */
	struct increq incr3b;                 /* incremental out3b values */
	struct scratch tmp;                   /* intermediate/constant values */
	mpfr_t vDi;                           /* variance of expected spacing */
	mpfr_t tot;                           /* cumulative out3a results */
	int ktrace;                           /* true to trace sub-terms */
	int sgn;                              /* add/subtract terms for cum */
	int ilo, ihi;                         /* index bounds for spacing */
	int klo, khi;                         /* dilog series index bounds */
	int i, k;
	int err;

	parse_commandline(argc, argv, &args);

	mpfr_init2(vDi, NBIT);
	mpfr_init2(tot, NBIT);

	if (ACT_BIST == args.action) {
		printf("running BIST\n");
		BIST_gcd();
		BIST_mpfr_gcd();
		BIST_remove_nonpos();
		BIST_simplify_factfrac();
		BIST_init_scratch();
		goto cleanup;
	} else {
		if ((2 <= args.i) && (args.i <= args.n)) {
			ilo = args.i;
			ihi = args.i;
		} else {
			ilo = 2;
			ihi = args.n;
		}

		for (i=ilo; i<=ihi; i++) {
			switch (args.action) {
			case ACT_VDI:
				var_Dlogistic(vDi, i, args.n, args.s, args.trace);
				break;
			case ACT_INNER:
				var_Dlogistic_inner(vDi, args.z, i, args.n, args.s, args.trace);
				break;
			default:
				fprintf(stderr, "unknown action");
				err = -1;
				goto cleanup;
			}
			printf("  %3d  %18.10g\n", i, mpfr_get_d(vDi, RND));
		}
	}

	err = 0;
cleanup:
	mpfr_clear(vDi);
	mpfr_clear(tot);
	return err;
}
